from typer import _click


def test_detect_program_name_submodule_path() -> None:
    class MainModule:
        __package__ = "example"

    program_name = _click.utils._detect_program_name(
        path="/tmp/cli.py",
        _main=MainModule(),
    )

    assert program_name == "python -m example.cli"



import subprocess
import sys

import typer
from typer.testing import CliRunner

runner = CliRunner()


def test_double_dash() -> None:
    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "coverage",
            "run",
            "-m",
            "typer",
            "tests/assets/cli/sample.py",
            "run",
            "hello",
            "--",
            "--name",
            "Camila",
        ],
        capture_output=True,
        encoding="utf-8",
    )
    assert "Got unexpected extra argument" in result.stderr
    assert "--name Camila" in result.stderr


def test_unknown_short_option() -> None:
    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "coverage",
            "run",
            "-m",
            "typer",
            "tests/assets/cli/sample.py",
            "run",
            "hello",
            "-x",
        ],
        capture_output=True,
        encoding="utf-8",
    )
    assert "No such option: -x" in result.stderr


def test_ignore_unknown_short_option() -> None:
    app = typer.Typer(
        context_settings={"ignore_unknown_options": True, "allow_extra_args": True}
    )

    @app.command()
    def main(
        ctx: typer.Context, all_: bool = typer.Option(False, "--all", "-a")
    ) -> None:
        assert all_
        print(ctx.args)

    result = runner.invoke(app, ["-azq"])
    assert result.exit_code == 0
    assert "['-zq']" in result.output



import os
import subprocess
import sys

from . import choice_case_insensitive_example as mod


def test_script() -> None:
    result = subprocess.run(
        [sys.executable, "-m", "coverage", "run", mod.__file__, "--name", "rick"],
        capture_output=True,
        encoding="utf-8",
    )
    assert result.returncode == 0
    assert "rick" in result.stdout


def test_completion_choice_bash_case_insensitive() -> None:
    result = subprocess.run(
        [sys.executable, "-m", "coverage", "run", mod.__file__, " "],
        capture_output=True,
        encoding="utf-8",
        env={
            **os.environ,
            "_CHOICE_CASE_INSENSITIVE_EXAMPLE.PY_COMPLETE": "complete_bash",
            "COMP_WORDS": "choice_case_insensitive_example.py --name MO",
            "COMP_CWORD": "2",
        },
    )
    assert result.returncode == 0
    assert "morty" in result.stdout
    assert "rick" not in result.stdout



import subprocess
import sys
import uuid

import typer
from typer.testing import CliRunner

from docs_src.parameter_types.uuid import tutorial001_py310 as mod

runner = CliRunner()
app = mod.app


def test_type_repr():
    command = typer.main.get_command(app)
    user_id_param = next(param for param in command.params if param.name == "user_id")
    assert repr(user_id_param.type) == "UUID"


def test_main():
    result = runner.invoke(app, ["d48edaa6-871a-4082-a196-4daab372d4a1"])
    assert result.exit_code == 0
    assert "User ID is d48edaa6-871a-4082-a196-4daab372d4a1" in result.output
    assert "UUID version is: 4" in result.output


def test_main_with_uuid_object():
    user_id = uuid.UUID("d48edaa6-871a-4082-a196-4daab372d4a1")
    result = runner.invoke(app, [], default_map={"user_id": user_id})
    assert result.exit_code == 0
    assert "User ID is d48edaa6-871a-4082-a196-4daab372d4a1" in result.output
    assert "UUID version is: 4" in result.output


def test_invalid_uuid():
    result = runner.invoke(app, ["7479706572-72756c6573"])
    assert result.exit_code != 0
    assert "Invalid value for 'user_id'" in result.output
    assert "not a valid UUID" in result.output


def test_script():
    result = subprocess.run(
        [sys.executable, "-m", "coverage", "run", mod.__file__, "--help"],
        capture_output=True,
        encoding="utf-8",
    )
    assert "Usage" in result.stdout



import typer

app = typer.Typer()


@app.command()
def visible():
    """Visible command."""


@app.command(hidden=True)
def hidden_decorated():
    """Hidden via @app.command(hidden=True)."""


def hidden_var():
    """Hidden via app.command(name, hidden=True)(fn)."""


app.command("hidden-var", hidden=True)(hidden_var)

hidden_subgroup = typer.Typer(hidden=True)


@hidden_subgroup.command()
def sub():
    """Hidden subgroup command."""


app.add_typer(hidden_subgroup, name="hidden-subgroup")



import os
import subprocess
import sys
import typing
from pathlib import Path
from typing import Annotated
from unittest import mock

import pytest
import typer
import typer._completion_shared
import typer.completion
from typer import _click
from typer.main import solve_typer_info_defaults, solve_typer_info_help
from typer.models import ParameterInfo, TyperInfo
from typer.testing import CliRunner

from .utils import requires_completion_permission

runner = CliRunner()


def test_help_from_info():
    # Mainly for coverage/completeness
    value = solve_typer_info_help(TyperInfo())
    assert value is None


def test_defaults_from_info():
    # Mainly for coverage/completeness
    value = solve_typer_info_defaults(TyperInfo())
    assert value


def test_too_many_parsers():
    def custom_parser(value: str) -> int:
        return int(value)  # pragma: no cover

    class CustomClickParser(_click.types.ParamType):
        name = "custom_parser"

        def convert(
            self,
            value: str,
            param: _click.Parameter | None,
            ctx: _click.Context | None,
        ) -> typing.Any:
            return int(value)  # pragma: no cover

    expected_error = (
        "Multiple custom type parsers provided. "
        "`parser` and `click_type` may not both be provided."
    )

    with pytest.raises(ValueError, match=expected_error):
        ParameterInfo(parser=custom_parser, click_type=CustomClickParser())


def test_valid_parser_permutations():
    def custom_parser(value: str) -> int:
        return int(value)  # pragma: no cover

    class CustomClickParser(_click.types.ParamType):
        name = "custom_parser"

        def convert(
            self,
            value: str,
            param: _click.Parameter | None,
            ctx: _click.Context | None,
        ) -> typing.Any:
            return int(value)  # pragma: no cover

    ParameterInfo()
    ParameterInfo(parser=custom_parser)
    ParameterInfo(click_type=CustomClickParser())


@requires_completion_permission
def test_install_invalid_shell():
    app = typer.Typer()

    @app.command()
    def main():
        print("Hello World")

    with mock.patch.object(
        typer._completion_shared, "_get_shell_name", return_value="xshell"
    ):
        result = runner.invoke(app, ["--install-completion"])
        assert "Shell xshell is not supported." in result.stdout
    result = runner.invoke(app)
    assert "Hello World" in result.stdout


def test_callback_too_many_parameters():
    app = typer.Typer()

    def name_callback(ctx, param, val1, val2):
        pass  # pragma: no cover

    @app.command()
    def main(name: str = typer.Option(..., callback=name_callback)):
        pass  # pragma: no cover

    with pytest.raises(_click.ClickException) as exc_info:
        runner.invoke(app, ["--name", "Camila"])
    assert (
        exc_info.value.message == "Too many CLI parameter callback function parameters"
    )


def test_callback_2_untyped_parameters():
    app = typer.Typer()

    def name_callback(ctx, value):
        print(f"info name is: {ctx.info_name}")
        print(f"value is: {value}")

    @app.command()
    def main(name: str = typer.Option(..., callback=name_callback)):
        print("Hello World")

    result = runner.invoke(app, ["--name", "Camila"])
    assert "info name is: main" in result.stdout
    assert "value is: Camila" in result.stdout


@pytest.mark.parametrize(
    ("param_hint", "option_decls", "expected_message"),
    [
        ("--name", (), "Invalid value for --name"),
        (None, ("--name", "-n"), "Invalid value for '--name' / '-n'"),
    ],
)
def test_bad_parameter_callback(
    param_hint: str | None, option_decls: tuple[str, ...], expected_message: str
) -> None:
    app = typer.Typer()

    def my_bad(value: str) -> str:
        kwargs = {"param_hint": param_hint} if param_hint is not None else {}
        raise typer.BadParameter("custom validation failed", **kwargs)

    @app.command()
    def main(name: str = typer.Option(..., *option_decls, callback=my_bad)) -> None:
        typer.echo(name)  # pragma: no cover

    result = runner.invoke(app, ["--name", "Camila"])
    assert result.exit_code == 2
    assert expected_message in result.stderr
    assert "custom validation failed" in result.stderr


def test_bad_parameter_main() -> None:
    app = typer.Typer()

    @app.command()
    def main() -> None:
        raise typer.BadParameter("custom validation failed")

    result = runner.invoke(app, [])
    assert result.exit_code == 2
    assert "Invalid value: custom validation failed" in result.stderr


@pytest.mark.parametrize(
    ("kw", "msg"),
    [
        (
            {"param_hint": ["--name", "-n"], "param_type": "parameter"},
            "Missing parameter '--name' / '-n'.",
        ),
        ({"param_type": "value"}, "Missing value."),
    ],
)
def test_missing_parameter_msg(kw: dict[str, object], msg: str) -> None:
    app = typer.Typer(rich_markup_mode=None)

    @app.command()
    def main() -> None:
        raise typer._click.exceptions.MissingParameter(**kw)

    result = runner.invoke(app, [])
    assert result.exit_code == 2
    assert msg in result.stderr


def test_missing_parameter_callback_msg() -> None:
    def my_cb(ctx: typer.Context, param: typer.CallbackParam, value: str) -> str:
        raise typer._click.exceptions.MissingParameter(
            message="My bad", ctx=ctx, param=param, param_type="parameter"
        )

    app = typer.Typer(rich_markup_mode=None)

    @app.command()
    def main(
        mode: Annotated[
            typing.Literal["alpha", "beta"],
            typer.Option(..., "--mode", callback=my_cb),
        ],
    ) -> None:
        typer.echo(mode)  # pragma: no cover

    result = runner.invoke(app, ["--mode", "alpha"])
    assert result.exit_code == 2
    assert "Missing parameter '--mode'." in result.stderr
    assert "My bad. Choose from:" in result.stderr
    assert "alpha" in result.stderr
    assert "beta" in result.stderr
    result_msg = runner.invoke(app, ["--mode", "alpha"], standalone_mode=False)
    assert isinstance(result_msg.exception, typer._click.exceptions.MissingParameter)
    assert str(result_msg.exception) == "My bad"


def test_missing_parameter_str() -> None:
    def my_cb(ctx: typer.Context, param: typer.CallbackParam, value: str) -> str:
        raise typer._click.exceptions.MissingParameter(ctx=ctx, param=param)

    app = typer.Typer()

    @app.command()
    def main(mode: str = typer.Option(..., "--mode", callback=my_cb)) -> None:
        typer.echo(mode)  # pragma: no cover

    result2 = runner.invoke(app, ["--mode", "alpha"], standalone_mode=False)
    assert isinstance(result2.exception, typer._click.exceptions.MissingParameter)
    assert str(result2.exception) == "Missing parameter: mode"


def test_click_exception_show_default_file() -> None:
    app = typer.Typer(rich_markup_mode=None)

    @app.command()
    def main() -> None:
        raise typer._click.ClickException("custom click failure")

    result = runner.invoke(app, [])
    assert result.exit_code == 1
    assert "custom click" in result.stderr
    assert "failure" in result.stderr


def test_no_args_is_help_show() -> None:
    app = typer.Typer(rich_markup_mode=None)

    @app.callback(invoke_without_command=True, no_args_is_help=True)
    def main() -> None:
        return None  # pragma: no cover

    result = runner.invoke(app, [])
    assert result.exit_code == 2
    assert "Usage:" in result.stderr
    assert "Show this message and exit." in result.stderr


def test_callback_3_untyped_parameters():
    app = typer.Typer()

    def name_callback(ctx, param, value):
        print(f"info name is: {ctx.info_name}")
        print(f"param name is: {param.name}")
        print(f"value is: {value}")

    @app.command()
    def main(name: str = typer.Option(..., callback=name_callback)):
        print("Hello World")

    result = runner.invoke(app, ["--name", "Camila"])
    assert "info name is: main" in result.stdout
    assert "param name is: name" in result.stdout
    assert "value is: Camila" in result.stdout


def test_callback_4_list_none():
    app = typer.Typer()

    def names_callback(ctx, param, values: list[str] | None):
        if values is None:
            return values
        return [value.upper() for value in values]

    @app.command()
    def main(
        names: list[str] | None = typer.Option(None, "--name", callback=names_callback),
    ):
        if names is None:
            print("Hello World")
        else:
            print(f"Hello {', '.join(names)}")

    result = runner.invoke(app, ["--name", "Sideshow", "--name", "Bob"])
    assert "Hello SIDESHOW, BOB" in result.stdout

    result = runner.invoke(app, [])
    assert "Hello World" in result.stdout


def test_multiple_bool_flags() -> None:
    app = typer.Typer()

    @app.command()
    def main(choices: list[bool] = typer.Option([], "--accept/--reject")) -> None:
        print(choices)

    result = runner.invoke(app, ["--accept", "--reject", "--accept"])
    assert result.exit_code == 0
    assert "[True, False, True]" in result.stdout


def test_empty_list_default_generator():
    def empty_list() -> list[str]:
        return []

    app = typer.Typer()

    @app.command()
    def main(
        names: Annotated[list[str], typer.Option(default_factory=empty_list)],
    ):
        print(names)

    result = runner.invoke(app)
    assert "[]" in result.output


def test_option_envvar():
    app = typer.Typer()

    @app.command()
    def main(user: Annotated[str, typer.Option(envvar="ME")]):
        print(f"Hello {user}")

    result = runner.invoke(app, env={"ME": "rick"})
    assert result.exit_code == 0
    assert "Hello rick" in result.output


def test_option_envvar_list():
    app = typer.Typer()

    @app.command()
    def main(users: Annotated[list[str], typer.Option(envvar="ME")]):
        for u in users:
            print(f"Hello {u}")

    result = runner.invoke(app, env={"ME": "rick morty"})
    assert result.exit_code == 0
    assert "Hello rick" in result.output
    assert "Hello morty" in result.output


def test_completion_argument():
    file_path = Path(__file__).parent / "assets/completion_argument.py"
    result = subprocess.run(
        [sys.executable, "-m", "coverage", "run", str(file_path), "E"],
        capture_output=True,
        encoding="utf-8",
        env={
            **os.environ,
            "_COMPLETION_ARGUMENT.PY_COMPLETE": "complete_zsh",
            "_TYPER_COMPLETE_ARGS": "completion_argument.py E",
            "_TYPER_COMPLETE_TESTING": "True",
        },
    )
    assert "Emma" in result.stdout or "_files" in result.stdout
    assert "ctx: completion_argument" in result.stderr
    assert "arg is: name" in result.stderr
    assert "incomplete is: E" in result.stderr


def test_completion_untyped_parameters():
    file_path = Path(__file__).parent / "assets/completion_no_types.py"
    result = subprocess.run(
        [sys.executable, "-m", "coverage", "run", str(file_path)],
        capture_output=True,
        encoding="utf-8",
        env={
            **os.environ,
            "_COMPLETION_NO_TYPES.PY_COMPLETE": "complete_zsh",
            "_TYPER_COMPLETE_ARGS": "completion_no_types.py --name Sebastian --name Ca",
        },
    )
    assert "info name is: completion_no_types.py" in result.stderr
    assert "args is: []" in result.stderr
    assert "incomplete is: Ca" in result.stderr
    assert '"Camila":"The reader of books."' in result.stdout
    assert '"Carlos":"The writer of scripts."' in result.stdout

    result = subprocess.run(
        [sys.executable, "-m", "coverage", "run", str(file_path)],
        capture_output=True,
        encoding="utf-8",
    )
    assert "Hello World" in result.stdout


def test_completion_untyped_parameters_different_order_correct_names():
    file_path = Path(__file__).parent / "assets/completion_no_types_order.py"
    result = subprocess.run(
        [sys.executable, "-m", "coverage", "run", str(file_path)],
        capture_output=True,
        encoding="utf-8",
        env={
            **os.environ,
            "_COMPLETION_NO_TYPES_ORDER.PY_COMPLETE": "complete_zsh",
            "_TYPER_COMPLETE_ARGS": "completion_no_types_order.py --name Sebastian --name Ca",
        },
    )
    assert "info name is: completion_no_types_order.py" in result.stderr
    assert "args is: []" in result.stderr
    assert "incomplete is: Ca" in result.stderr
    assert '"Camila":"The reader of books."' in result.stdout
    assert '"Carlos":"The writer of scripts."' in result.stdout

    result = subprocess.run(
        [sys.executable, "-m", "coverage", "run", str(file_path)],
        capture_output=True,
        encoding="utf-8",
    )
    assert "Hello World" in result.stdout


def test_autocompletion_too_many_parameters():
    app = typer.Typer()

    def name_callback(ctx, args, incomplete, val2):
        pass  # pragma: no cover

    @app.command()
    def main(name: str = typer.Option(..., autocompletion=name_callback)):
        pass  # pragma: no cover

    with pytest.raises(_click.ClickException) as exc_info:
        runner.invoke(app, ["--name", "Camila"])
    assert exc_info.value.message == "Invalid autocompletion callback parameters: val2"


def test_forward_references():
    app = typer.Typer()

    @app.command()
    def main(arg1, arg2: int, arg3: "int", arg4: bool = False, arg5: "bool" = False):
        print(f"arg1: {type(arg1)} {arg1}")
        print(f"arg2: {type(arg2)} {arg2}")
        print(f"arg3: {type(arg3)} {arg3}")
        print(f"arg4: {type(arg4)} {arg4}")
        print(f"arg5: {type(arg5)} {arg5}")

    result = runner.invoke(app, ["Hello", "2", "invalid"])

    assert "Invalid value for 'arg3'" in result.output
    assert "'invalid' is not a valid int" in result.output
    result = runner.invoke(app, ["Hello", "2", "3", "--arg4", "--arg5"])
    assert (
        "arg1: <class 'str'> Hello\narg2: <class 'int'> 2\narg3: <class 'int'> 3\narg4: <class 'bool'> True\narg5: <class 'bool'> True\n"
        in result.stdout
    )


def test_context_settings_inheritance_single_command():
    app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]})

    @app.command()
    def main(name: str):
        pass  # pragma: no cover

    result = runner.invoke(app, ["main", "-h"])
    assert "Show this message and exit." in result.stdout


def test_options_metadata_typer_default():
    app = typer.Typer(options_metavar="[options]")

    @app.command()
    def c1():
        pass  # pragma: no cover

    @app.command(options_metavar="[OPTS]")
    def c2():
        pass  # pragma: no cover

    result = runner.invoke(app, ["c1", "--help"])
    assert "Usage: root c1 [options]" in result.stdout

    result = runner.invoke(app, ["c2", "--help"])
    assert "Usage: root c2 [OPTS]" in result.stdout



"""
Tests for the Windows console functionality.
Created after vendoring Click to ensure test coverage is back up to 100%.
"""

import ctypes
import io
import sys

import pytest
import typer
from typer.testing import CliRunner

from .utils import needs_windows

pytestmark = needs_windows


if sys.platform == "win32":
    from typer._click import _compat, _winconsole


def _identity_buffer(obj, writable=False):  # noqa: ARG001
    return obj


def _route_console_stream(target_name, wrapper, state=None):
    def patched_windows_stream(stream, encoding, errors):  # noqa: ARG001
        current_target = getattr(sys, target_name)
        if stream is current_target:
            if state is not None and target_name == "stderr":
                state["stderr_wrap_calls"] += 1
            buffer = getattr(stream, "buffer", None)
            return wrapper(buffer) if buffer else None
        return None

    return patched_windows_stream


def _capture_write_console(state):
    def fake_write_console(handle, buffer, units_to_write, units_written_ptr, reserved):  # noqa: ARG001
        state["write_calls"] += 1
        bytes_to_write = units_to_write * 2
        state["written"].extend(buffer[:bytes_to_write])
        units_written_ptr._obj.value = units_to_write
        return 1

    return fake_write_console


def test_winconsole_stdin(monkeypatch):
    runner = CliRunner()
    app = typer.Typer()

    @app.command()
    def read_name(config: typer.FileText = typer.Option(...)) -> None:
        name = config.readline().strip()
        typer.echo(f"Hello {name}")

    utf16_data = bytearray("Rick\r\n".encode("utf-16-le"))
    state = {"pos": 0, "read_calls": 0}

    def fake_read_console(handle, buffer, units_to_read, units_read_ptr, reserved):  # noqa: ARG001
        state["read_calls"] += 1
        max_bytes = units_to_read * 2
        chunk = utf16_data[state["pos"] : state["pos"] + max_bytes]
        if chunk:
            buffer[0 : len(chunk)] = chunk
            state["pos"] += len(chunk)
            units_read_ptr._obj.value = len(chunk) // 2
            return 1

        return 1  # pragma: no cover

    monkeypatch.setattr(_winconsole, "get_buffer", _identity_buffer)
    monkeypatch.setattr(_winconsole, "ReadConsoleW", fake_read_console)
    monkeypatch.setattr(_winconsole, "GetLastError", lambda: 0)
    monkeypatch.setattr(
        _compat,
        "_get_windows_console_stream",
        _route_console_stream("stdin", _winconsole._get_text_stdin),
    )

    result = runner.invoke(app, ["--config", "-"])
    assert result.exit_code == 0, result.output
    assert "Hello Rick" in result.stdout
    assert state["read_calls"] > 0


def test_winconsole_stdout(monkeypatch):
    runner = CliRunner()
    app = typer.Typer()
    state = {"write_calls": 0, "written": bytearray()}

    @app.command()
    def write_message(out: typer.FileTextWrite = typer.Option(...)) -> None:
        out.write("Hello Summer\n")

    monkeypatch.setattr(_winconsole, "get_buffer", _identity_buffer)
    monkeypatch.setattr(_winconsole, "WriteConsoleW", _capture_write_console(state))
    monkeypatch.setattr(_winconsole, "GetLastError", lambda: 0)
    monkeypatch.setattr(
        _compat,
        "_get_windows_console_stream",
        _route_console_stream("stdout", _winconsole._get_text_stdout),
    )

    result = runner.invoke(app, ["--out", "-"])
    assert result.exit_code == 0, result.output
    assert state["write_calls"] > 0
    assert _winconsole._WindowsConsoleWriter(1).isatty() is True
    decoded = state["written"].decode("utf-16-le", errors="ignore")
    assert "Hello Summer\r\n" in decoded


def test_winconsole_stderr(monkeypatch):
    runner = CliRunner()
    app = typer.Typer()
    state = {"write_calls": 0, "written": bytearray(), "stderr_wrap_calls": 0}

    @app.command()
    def main() -> None:
        typer.echo("Ran out of adventure time!", err=True)

    monkeypatch.setattr(_winconsole, "get_buffer", _identity_buffer)
    monkeypatch.setattr(_winconsole, "WriteConsoleW", _capture_write_console(state))
    monkeypatch.setattr(_winconsole, "GetLastError", lambda: 0)
    monkeypatch.setattr(
        _compat,
        "_get_windows_console_stream",
        _route_console_stream("stderr", _winconsole._get_text_stderr, state),
    )

    result = runner.invoke(app)
    assert result.exit_code == 0, result.output
    assert state["stderr_wrap_calls"] > 0
    assert state["write_calls"] > 0
    decoded = state["written"].decode("utf-16-le", errors="ignore")
    assert "Ran out of adventure time!\r\n" in decoded


@pytest.mark.parametrize(
    ("writable", "source", "expected_flags"),
    [
        (True, bytearray(b"python"), 1),  # PyBUF_WRITABLE
        (False, b"python", 0),  # PyBUF_SIMPLE
    ],
)
def test_get_buffer(monkeypatch, writable, source, expected_flags):
    state = {"flags": None, "released": 0}
    if writable:
        backing = (_winconsole.c_char * len(source)).from_buffer(source)
    else:
        backing = (_winconsole.c_char * len(source)).from_buffer_copy(source)
    backing_ptr = _winconsole.c_void_p(ctypes.addressof(backing))

    def fake_object_get_buffer(obj, buf_ref, flags):  # noqa: ARG001
        state["flags"] = flags
        buf = buf_ref._obj
        buf.buf = backing_ptr
        buf.len = len(source)

    def fake_buffer_release(buf_ref):  # noqa: ARG001
        state["released"] += 1

    monkeypatch.setattr(_winconsole, "PyObject_GetBuffer", fake_object_get_buffer)
    monkeypatch.setattr(_winconsole, "PyBuffer_Release", fake_buffer_release)

    probe = source if writable else b"x"
    out = _winconsole.get_buffer(probe, writable=writable)
    if writable:
        # mutate the first byte of "python" to obtain another beloved programming language
        out[0] = b"c"
        assert source == bytearray(b"cython")
    else:
        assert bytes(out[: len(source)]) == source
    assert state["flags"] == expected_flags
    assert state["released"] == 1


def test_isatty():
    assert _winconsole._WindowsConsoleRawIOBase(None).isatty() is True
    assert _winconsole._WindowsConsoleReader(0).isatty() is True
    assert _winconsole._WindowsConsoleReader(1).isatty() is True


def test_console_stream():
    class NamedBytesIO(io.BytesIO):
        name = "fake-buffer"

        def isatty(self):
            return False

    stream = _winconsole.ConsoleStream(
        io.TextIOWrapper(io.BytesIO(), encoding="utf-8"), NamedBytesIO()
    )
    assert stream.isatty() is False
    assert stream.name == "fake-buffer"
    assert "fake-buffer" in repr(stream)
    assert "utf-8" in repr(stream)

    # test writelines
    stream.writelines(["hello", " ", "world"])
    stream._text_stream.flush()
    assert stream._text_stream.buffer.getvalue().decode("utf-8") == "hello world"

    # Cover bytes write path.
    assert stream.write(b"!") == 1
    assert stream.buffer.getvalue().endswith(b"!")


@pytest.mark.parametrize(
    ("error", "msg"),
    [
        (0, "ERROR_SUCCESS"),  # ERROR_SUCCESS
        (8, "ERROR_NOT_ENOUGH_MEMORY"),  # ERROR_NOT_ENOUGH_MEMORY
        (342, "Windows error 342"),
    ],
)
def test_error_message(error, msg):
    writer = _winconsole._WindowsConsoleWriter
    assert writer._get_error_message(error) == msg


def test_is_console():
    assert _winconsole._is_console(object()) is False


def test_get_windows_console_stream_factory_and_buffer_paths(monkeypatch):
    monkeypatch.setattr(_winconsole, "_is_console", lambda f: True)
    monkeypatch.setattr(_winconsole, "get_buffer", object())

    class FakeStream:
        def __init__(self, fd, buffer=None):
            self._fd = fd
            self.buffer = buffer

        def fileno(self):
            return self._fd

    wrapped = {"called": False, "buffer": None}

    def fake_factory(buffer):
        wrapped["called"] = True
        wrapped["buffer"] = buffer
        return "wrapped", buffer

    monkeypatch.setattr(_winconsole, "_stream_factories", {7: fake_factory})

    # Known console stream preconditions pass, but no stream factory for this fd.
    get_stream = _winconsole._get_windows_console_stream
    assert get_stream(FakeStream(99, object()), "utf-16-le", "strict") is None

    # Factory exists, but stream has no usable .buffer.
    assert get_stream(FakeStream(7, None), "utf-16-le", "strict") is None

    # Factory exists and buffer is present, so wrapper result is returned.
    raw_buffer = object()
    out = get_stream(FakeStream(7, raw_buffer), "utf-16-le", "strict")
    assert out == ("wrapped", raw_buffer)
    assert wrapped["called"] is True
    assert wrapped["buffer"] is raw_buffer


def test_windows_console_reader(monkeypatch):
    reader = _winconsole._WindowsConsoleReader(42)

    # Empty input buffer returns early
    assert reader.readinto(bytearray()) == 0

    # Require an even number of bytes
    with pytest.raises(ValueError):
        reader.readinto(bytearray(3))

    def writable_buffer(obj, writable=False):  # noqa: ARG001
        return (ctypes.c_char * len(obj)).from_buffer(obj)

    monkeypatch.setattr(_winconsole, "get_buffer", writable_buffer)

    def patch_console(read_console, error):
        monkeypatch.setattr(_winconsole, "ReadConsoleW", read_console)
        monkeypatch.setattr(_winconsole, "GetLastError", lambda: error)

    def make_read(payload=b"", rv=1, units_read=None):
        def read_console(handle, buffer, units_to_read, units_read_ptr, reserved):  # noqa: ARG001
            bytes_to_copy = min(len(payload), units_to_read * 2)
            if bytes_to_copy:
                buffer[0:bytes_to_copy] = payload[:bytes_to_copy]
            read_units = units_read if units_read is not None else bytes_to_copy // 2
            units_read_ptr._obj.value = read_units
            return rv

        return read_console

    # Normal successful read returns the number of bytes read
    patch_console(make_read(payload=b"A\x00B\x00"), _winconsole.ERROR_SUCCESS)
    assert reader.readinto(bytearray(4)) == 4

    # CTRL+Z (EOF) should be translated into an empty read
    patch_console(
        make_read(payload=_winconsole.EOF + b"\x00", units_read=1),
        _winconsole.ERROR_SUCCESS,
    )
    assert reader.readinto(bytearray(2)) == 0

    # An aborted read should sleep briefly while waiting for KeyboardInterrupt
    sleep_state = {"calls": 0}

    def fake_sleep(seconds):
        sleep_state["calls"] += 1
        assert seconds == 0.1

    monkeypatch.setattr(
        _winconsole, "time", type("FakeTime", (), {"sleep": fake_sleep})
    )
    patch_console(
        make_read(payload=b"Z\x00", units_read=1),
        _winconsole.ERROR_OPERATION_ABORTED,
    )
    assert reader.readinto(bytearray(2)) == 2
    assert sleep_state["calls"] == 1

    # Failed reads propagate a Windows error
    patch_console(make_read(rv=0), _winconsole.ERROR_NOT_ENOUGH_MEMORY)
    with pytest.raises(OSError, match="Windows error"):
        reader.readinto(bytearray(2))



import os
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Literal

import pytest
import typer
from typer import _click, models
from typer.testing import CliRunner

from tests.utils import needs_linux, needs_windows

runner = CliRunner()


def test_optional():
    app = typer.Typer()

    @app.command()
    def opt(user: str | None = None):
        if user:
            print(f"User: {user}")
        else:
            print("No user")

    result = runner.invoke(app)
    assert result.exit_code == 0
    assert "No user" in result.output

    result = runner.invoke(app, ["--user", "Camila"])
    assert result.exit_code == 0
    assert "User: Camila" in result.output


def test_union_type_optional():
    app = typer.Typer()

    @app.command()
    def opt(user: str | None = None):
        if user:
            print(f"User: {user}")
        else:
            print("No user")

    result = runner.invoke(app)
    assert result.exit_code == 0
    assert "No user" in result.output

    result = runner.invoke(app, ["--user", "Camila"])
    assert result.exit_code == 0
    assert "User: Camila" in result.output


def test_optional_tuple():
    app = typer.Typer()

    @app.command()
    def opt(number: tuple[int, int] | None = None):
        if number:
            print(f"Number: {number}")
        else:
            print("No number")

    result = runner.invoke(app)
    assert result.exit_code == 0
    assert "No number" in result.output

    result = runner.invoke(app, ["--number", "4", "2"])
    assert result.exit_code == 0
    assert "Number: (4, 2)" in result.output


def test_no_type():
    app = typer.Typer()

    @app.command()
    def no_type(user):
        print(f"User: {user}")

    result = runner.invoke(app, ["Camila"])
    assert result.exit_code == 0
    assert "User: Camila" in result.output


class SomeEnum(Enum):
    ONE = "one"
    TWO = "two"
    THREE = "three"


@pytest.mark.parametrize(
    "type_annotation",
    [list[Path], list[SomeEnum], list[str]],
)
def test_list_parameters_convert_to_lists(type_annotation):
    # Lists containing objects that are converted by Click (i.e. not Path or Enum)
    # should not be inadvertently converted to tuples
    expected_element_type = type_annotation.__args__[0]
    app = typer.Typer()

    @app.command()
    def list_conversion(container: type_annotation):
        assert isinstance(container, list)
        for element in container:
            assert isinstance(element, expected_element_type)

    result = runner.invoke(app, ["one", "two", "three"])
    assert result.exit_code == 0


@pytest.mark.parametrize(
    "type_annotation",
    [
        tuple[str, str],
        tuple[str, Path],
        tuple[Path, Path],
        tuple[str, SomeEnum],
        tuple[SomeEnum, SomeEnum],
    ],
)
def test_tuple_parameter_elements_are_converted_recursively(type_annotation):
    # Tuple elements that aren't converted by Click (i.e. Path or Enum)
    # should be recursively converted by Typer
    expected_element_types = type_annotation.__args__
    app = typer.Typer()

    @app.command()
    def tuple_recursive_conversion(container: type_annotation):
        assert isinstance(container, tuple)
        for element, expected_type in zip(
            container, expected_element_types, strict=True
        ):
            assert isinstance(element, expected_type)

    result = runner.invoke(app, ["one", "two"])
    assert result.exit_code == 0


def test_tuple_wrong_arity():
    app = typer.Typer()

    @app.command()
    def tuple_arity(value: tuple[str, str] = typer.Option(...)):
        print(value)  # pragma: no cover

    result = runner.invoke(app, [], default_map={"value": ("only-one",)})
    assert result.exit_code == 2
    assert "2 values are required, but 1 given." in result.output


def test_custom_parse():
    app = typer.Typer()

    @app.command()
    def custom_parser(
        hex_value: int = typer.Argument(None, parser=lambda x: int(x, 0)),
    ):
        assert hex_value == 0x56

    result = runner.invoke(app, ["0x56"])
    assert result.exit_code == 0


def test_custom_parse_value_error():
    app = typer.Typer()

    @app.command()
    def custom_parser(
        hex_value: int = typer.Argument(None, parser=lambda x: int(x, 0)),
    ):
        print(hex_value)  # pragma: no cover

    result = runner.invoke(app, ["not-a-hex"])
    assert result.exit_code == 2
    assert "Invalid value" in result.output


def test_custom_click_type():
    class BaseNumberParamType(_click.types.ParamType):
        name = "base_integer"

        def convert(
            self,
            value: Any,
            param: _click.Parameter | None,
            ctx: _click.Context | None,
        ) -> Any:
            return int(value, 0)

    app = typer.Typer()

    @app.command()
    def custom_click_type(
        hex_value: int = typer.Argument(None, click_type=BaseNumberParamType()),
    ):
        assert hex_value == 0x56

    result = runner.invoke(app, ["0x56"])
    assert result.exit_code == 0


def test_int_range_open_bound_clamp():
    app = typer.Typer()

    @app.command()
    def custom_click_type(
        value: int = typer.Argument(
            ...,
            click_type=_click.types.IntRange(min=1, min_open=True, clamp=True),
        ),
    ):
        print(value)

    result = runner.invoke(app, ["1"])
    assert result.exit_code == 0
    assert "2" in result.output


def test_bool_convert_invalid():
    app = typer.Typer()

    @app.command()
    def main(value: bool):
        print(value)  # pragma: no cover

    result = runner.invoke(app, ["maybe"])
    assert result.exit_code == 2
    assert "is not a valid boolean" in result.output
    assert "yes" in result.output
    assert "false" in result.output


@pytest.mark.parametrize(
    ("arg_enc", "system_enc", "raw_value", "expected_output"),
    [
        pytest.param("latin-1", "utf-8", b"\xff", "ÿ"),
        pytest.param("ascii", "latin-1", b"\xff", "ÿ"),
        pytest.param("ascii", "utf-16", b"\xff", "�"),
        pytest.param("ascii", "ascii", b"\xff", "�"),
    ],
)
def test_string_param_type_converts_bytes(
    monkeypatch: pytest.MonkeyPatch,
    arg_enc: str,
    system_enc: str,
    raw_value: bytes,
    expected_output: str,
):
    app = typer.Typer()

    @app.command()
    def show(name: str = typer.Option(...)):
        print(name)

    command = typer.main.get_command(app)
    name_param = next(param for param in command.params if param.name == "name")
    assert repr(name_param.type) == "STRING"

    monkeypatch.setattr(_click.types, "_get_argv_encoding", lambda: arg_enc)
    monkeypatch.setattr(_click.types.sys, "getfilesystemencoding", lambda: system_enc)

    result = runner.invoke(app, [], default_map={"name": raw_value})
    assert result.exit_code == 0
    assert expected_output in result.output


@pytest.mark.parametrize("path_type", [str, bytes, Path])
def test_path_coerced(path_type) -> None:
    # Ensure coerce_path_result works correctly
    app = typer.Typer()

    @app.command()
    def show(path: Any = typer.Option(..., path_type=path_type)):
        print(path)

    result = runner.invoke(app, ["--path", "dir/my_awesome_file.txt"])
    assert result.exit_code == 0
    assert "my_awesome_file" in result.output


@pytest.mark.parametrize(
    ("create_file", "option_kwargs", "deny_mode", "expected_error"),
    [
        (True, {"file_okay": False, "dir_okay": True}, None, "is a file"),
        (False, {"file_okay": True, "dir_okay": False}, None, "is a directory"),
        (True, {"readable": True}, os.R_OK, "is not readable"),
        (True, {"readable": False, "writable": True}, os.W_OK, "is not writable"),
    ],
)
def test_path_convert_failures(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    create_file: bool,
    option_kwargs: dict[str, bool],
    deny_mode: int | None,
    expected_error: str,
) -> None:
    app = typer.Typer()

    @app.command()
    def show(path: Path = typer.Option(..., **option_kwargs)):
        print(path)  # pragma: no cover

    if deny_mode is not None:
        original_access = os.access

        def fake_access(path: str, mode: int) -> bool:
            if mode == deny_mode:
                return False
            return original_access(path, mode)  # pragma: no cover

        monkeypatch.setattr(models.os, "access", fake_access)

    path = tmp_path / "some_path"
    if create_file:
        path.write_text("hello")
    else:
        path.mkdir()
    result = runner.invoke(app, ["--path", str(path)])

    assert result.exit_code != 0
    assert expected_error in result.output


def test_convert_type():
    from typer._click.types import convert_type

    # str
    assert convert_type(str) is _click.types.STRING
    assert convert_type(None) is _click.types.STRING
    assert convert_type(None, default=["a"]) is _click.types.STRING

    # tuples
    tuple_type = convert_type((str, int))
    assert isinstance(tuple_type, _click.types.Tuple)
    assert [type(item) for item in tuple_type.types] == [
        type(_click.types.STRING),
        type(_click.types.INT),
    ]

    guessed_tuple = convert_type(None, default=[(1, "x")])
    assert isinstance(guessed_tuple, _click.types.Tuple)
    assert [type(item) for item in guessed_tuple.types] == [
        type(_click.types.INT),
        type(_click.types.STRING),
    ]

    # numbers
    assert convert_type(int) is _click.types.INT
    assert convert_type(float) is _click.types.FLOAT
    assert convert_type(bool) is _click.types.BOOL

    param_type = _click.types.IntRange(min=0, max=10)
    assert convert_type(param_type) is param_type

    guessed_int = convert_type(None, default=42)
    assert guessed_int is _click.types.INT

    # custom type
    class CustomType:
        pass

    guessed_unknown = convert_type(None, default=CustomType())
    assert guessed_unknown is _click.types.STRING

    func_type = convert_type(CustomType)
    assert isinstance(func_type, _click.types.FuncParamType)
    assert func_type.name == "CustomType"


@pytest.mark.parametrize(
    ("platform_case", "stdin_encoding", "filesystem_encoding"),
    [
        pytest.param("windows", None, "utf-8", marks=needs_windows),
        pytest.param("linux", "latin-1", "utf-8", marks=needs_linux),
        pytest.param("linux", None, "latin-1", marks=needs_linux),
    ],
)
def test_argv_encoding(
    monkeypatch: pytest.MonkeyPatch,
    platform_case: str,
    stdin_encoding: str | None,
    filesystem_encoding: str,
) -> None:
    sys = _click._compat.sys
    if platform_case == "windows":
        import locale

        monkeypatch.setattr(locale, "getpreferredencoding", lambda: "latin-1")
    else:

        class FakeStdin:
            def __init__(self, encoding: str | None) -> None:
                self.encoding = encoding

        monkeypatch.setattr(sys, "stdin", FakeStdin(stdin_encoding))
        monkeypatch.setattr(sys, "getfilesystemencoding", lambda: filesystem_encoding)

    converted = _click.types.STRING.convert(b"\xff", None, None)
    assert converted == "ÿ"


@pytest.mark.parametrize(
    ("parameter", "expected_metavar"),
    [
        pytest.param(Annotated[str, typer.Option(...)], "<str>"),
        pytest.param(Annotated[str, typer.Argument(...)], "<str>"),
        pytest.param(Annotated[int, typer.Option(...)], "<int>"),
        pytest.param(Annotated[int, typer.Argument(...)], "<int>"),
        pytest.param(Annotated[float, typer.Option(...)], "<float>"),
        pytest.param(
            Annotated[float, typer.Option(..., min=0.666, max=3.42)], "<float range>"
        ),
        pytest.param(Annotated[tuple[str, int], typer.Option(...)], "<str int>..."),
        pytest.param(Annotated[tuple[Path, str], typer.Option(...)], "<path str>..."),
        pytest.param(Annotated[str, typer.Option(..., resolve_path=True)], "<str>"),
        pytest.param(Annotated[Path, typer.Option(...)], "<path>"),
        pytest.param(Annotated[Path, typer.Option(..., dir_okay=False)], "<file>"),
        pytest.param(
            Annotated[Path, typer.Option(..., file_okay=False)], "<directory>"
        ),
        pytest.param(Annotated[SomeEnum, typer.Option(...)], "<one|two|three>"),
        pytest.param(Annotated[SomeEnum, typer.Argument()], "<one|two|three>"),
        pytest.param(Annotated[Literal["x", "y"], typer.Option(...)], "<x|y>"),
        pytest.param(
            Annotated[datetime, typer.Option(..., formats=["%Y-%m-%d", "%d/%m/%Y"])],
            "<%Y-%m-%d|%d/%m/%Y>",
        ),
    ],
)
def test_param_type_help_metavar(parameter: Any, expected_metavar: str) -> None:
    app = typer.Typer()

    @app.command()
    # TODO: type-specific default
    def with_default(value: parameter = "my_default"):
        pass  # pragma: no cover

    @app.command()
    def without_default(value: parameter):
        pass  # pragma: no cover

    result = runner.invoke(app, ["with-default", "--help"])
    assert result.exit_code == 0
    assert expected_metavar in result.output

    result = runner.invoke(app, ["without-default", "--help"])
    assert result.exit_code == 0
    assert expected_metavar in result.output


def test_metavar_preformatted_square_brackets() -> None:
    app = typer.Typer()

    @app.command()
    def main(value: Annotated[str, typer.Argument(metavar="[VALUE]")]):
        pass  # pragma: no cover

    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "[VALUE]" in result.output
    assert "[[VALUE]]" not in result.output



import os
import subprocess
import sys

from . import file_example as mod


def test_script() -> None:
    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "coverage",
            "run",
            mod.__file__,
            "--config",
            mod.__file__,
        ],
        capture_output=True,
        encoding="utf-8",
    )
    assert result.returncode == 0
    assert "def main(config: typer.FileText = typer.Option(...)):" in result.stdout


def test_completion_file_bash() -> None:
    result = subprocess.run(
        [sys.executable, "-m", "coverage", "run", mod.__file__, " "],
        capture_output=True,
        encoding="utf-8",
        env={
            **os.environ,
            "_FILE_EXAMPLE.PY_COMPLETE": "complete_bash",
            "COMP_WORDS": "file_example.py --config file_ex",
            "COMP_CWORD": "2",
        },
    )
    assert result.returncode == 0
    assert "file_ex" in result.stdout



from enum import Enum

import pytest
import typer
from typer import _click
from typer.testing import CliRunner

app = typer.Typer(context_settings={"token_normalize_func": str.lower})


class User(str, Enum):
    rick = "Rick"
    morty = "Morty"


@app.command()
def hello_option(name: User = User.rick) -> None:
    print(f"Hello {name.value}!")


@app.command()
def hello_argument(name: User) -> None:
    print(f"Hello {name.value}!")


@app.command()
def hello_no_choices(
    name: User = typer.Option(..., "--name", show_choices=False),
):
    print(f"Hello {name.value}!")


@app.command()
def hello_all_options(name: list[str] = typer.Option(["World"])) -> None:
    for n in name:
        print(f"Hello {n}!")


@app.command()
def hello_all_args(names: list[str] = typer.Argument(["World"])) -> None:
    for name in names:
        print(f"Hello {name}!")


@app.command()
def hello_all_envvar(
    names: list[str] = typer.Argument(["World"], envvar="NAMES"),
) -> None:
    for name in names:
        print(f"Hello {name}!")


@app.command()
def split_variadic_and_pair(items: list[str], pair: tuple[str, str]) -> None:
    print(f"items={items}")
    print(f"pair={pair}")


runner = CliRunner()


def test_enum_choice() -> None:
    result = runner.invoke(
        app, ["hello-option", "--name", "morty"], catch_exceptions=False
    )
    assert result.exit_code == 0
    assert "Hello Morty!" in result.output

    result = runner.invoke(app, ["hello-option", "--name", "Rick"])
    assert result.exit_code == 0
    assert "Hello Rick!" in result.output

    result = runner.invoke(app, ["hello-option", "--name", "RICK"])
    assert result.exit_code == 0
    assert "Hello Rick!" in result.output

    result = runner.invoke(app, ["hello-no-choices", "--name", "RICK"])
    assert result.exit_code == 0
    assert "Hello Rick!" in result.output

    result = runner.invoke(app, ["hello-argument", "RICK"])
    assert result.exit_code == 0
    assert "Hello Rick!" in result.output


def test_enum_choice_repr() -> None:
    root_command = typer.main.get_command(app)
    command = root_command.commands["hello-option"]
    name_param = next(param for param in command.params if param.name == "name")
    assert repr(name_param.type).startswith("Choice([")


def test_enum_choice_help() -> None:
    result = runner.invoke(app, ["hello-argument", "--help"])
    assert result.exit_code == 0
    assert "<rick|morty>" in result.output

    result = runner.invoke(app, ["hello-option", "--help"])
    assert result.exit_code == 0
    assert "<rick|morty>" in result.output

    result = runner.invoke(app, ["hello-no-choices", "--help"])
    assert result.exit_code == 0
    assert "--name" in result.output
    assert "rick|morty" not in result.output


def test_enum_choice_missing_message() -> None:
    result = runner.invoke(app, ["hello-argument"])
    assert result.exit_code != 0
    assert "Missing argument" in result.output
    assert "Choose from:" in result.output
    assert "rick" in result.output
    assert "morty" in result.output


def test_list() -> None:
    result = runner.invoke(
        app, ["hello-all-options", "--name", "Rick", "--name", "Morty"]
    )
    assert result.exit_code == 0
    assert "Hello World!" not in result.output
    assert "Hello Rick!" in result.output
    assert "Hello Morty!" in result.output

    result = runner.invoke(app, ["hello-all-args", "Rick", "Morty"])
    assert result.exit_code == 0
    assert "Hello World!" not in result.output
    assert "Hello Rick!" in result.output
    assert "Hello Morty!" in result.output


def test_list_empty() -> None:
    result = runner.invoke(app, ["hello-all-args"])
    assert result.exit_code == 0
    assert "Hello World!" in result.output

    result = runner.invoke(app, ["hello-all-options"])
    assert result.exit_code == 0
    assert "Hello World!" in result.output

    result = runner.invoke(app, ["hello-all-envvar"])
    assert result.exit_code == 0
    assert "Hello World!" in result.output


def test_split_envvar_value(monkeypatch) -> None:
    # This will use split_envvar_value to produce two strings from the envvar
    monkeypatch.setenv("NAMES", "Rick   Morty")
    result = runner.invoke(app, ["hello-all-envvar"])
    assert result.exit_code == 0
    assert "Hello Rick!" in result.output
    assert "Hello Morty!" in result.output


def test_list_pair() -> None:
    result = runner.invoke(app, ["split-variadic-and-pair", "a", "b", "c", "x", "y"])
    assert result.exit_code == 0
    assert "items=['a', 'b', 'c']" in result.output
    assert "pair=('x', 'y')" in result.output


def test_float_range_open_bounds_with_clamp_not_allowed():
    with pytest.raises(TypeError, match="Clamping is not supported for open bounds."):
        _click.types.FloatRange(min=0.0, min_open=True, clamp=True)



from datetime import date
from pathlib import Path

import pytest
from typer.testing import CliRunner

from scripts.prepare_release import (
    BumpType,
    app,
    bump_version,
    get_release_notes_body,
    update_release_notes,
    update_version_file,
)

runner = CliRunner()


@pytest.mark.parametrize(
    ("current_version", "bump", "new_version"),
    [
        ("0.26.2", "major", "1.0.0"),
        ("0.26.2", "minor", "0.27.0"),
        ("0.26.2", "patch", "0.26.3"),
    ],
)
def test_bump_version(current_version: str, bump: BumpType, new_version: str) -> None:
    assert bump_version(current_version, bump) == new_version


def test_update_version_file() -> None:
    content = '"""Typer."""\n\n__version__ = "0.26.2"\n'

    new_content = update_version_file(content, "0.26.3", Path("typer/__init__.py"))

    assert new_content == '"""Typer."""\n\n__version__ = "0.26.3"\n'


def test_update_version_file_requires_newer_version() -> None:
    content = '__version__ = "0.26.2"\n'

    with pytest.raises(RuntimeError, match="must be greater"):
        update_version_file(content, "0.26.2", Path("typer/__init__.py"))


def test_update_release_notes() -> None:
    content = """# Release Notes

## Latest Changes

### Fixes

* Fix something.

## 0.26.2 (2026-05-27)

### Fixes

* Previous fix.
"""

    new_content = update_release_notes(
        content, "0.26.3", date(2026, 5, 28), Path("docs/release-notes.md")
    )

    assert (
        new_content
        == """# Release Notes

## Latest Changes

## 0.26.3 (2026-05-28)

### Fixes

* Fix something.

## 0.26.2 (2026-05-27)

### Fixes

* Previous fix.
"""
    )


def test_update_release_notes_rejects_existing_version() -> None:
    content = """# Release Notes

## Latest Changes

## 0.26.3 (2026-05-28)
"""

    with pytest.raises(RuntimeError, match="already contain"):
        update_release_notes(
            content, "0.26.3", date(2026, 5, 28), Path("docs/release-notes.md")
        )


def test_get_release_notes_body_with_dated_heading() -> None:
    content = """# Release Notes

## Latest Changes

## 0.26.3 (2026-05-28)

### Fixes

* Fix something.

## 0.26.2 (2026-05-27)

### Fixes

* Previous fix.
"""

    body = get_release_notes_body(content, "0.26.3", Path("docs/release-notes.md"))

    assert (
        body
        == """### Fixes

* Fix something.
"""
    )


def test_get_release_notes_body_with_plain_heading() -> None:
    content = """# Release Notes

## Latest Changes

## 0.26.3

### Fixes

* Fix something.
"""

    body = get_release_notes_body(content, "0.26.3", Path("docs/release-notes.md"))

    assert body == "### Fixes\n\n* Fix something.\n"


def test_get_release_notes_body_allows_non_version_h2_content() -> None:
    content = """# Release Notes

## Latest Changes

## 0.26.3

## Highlights

* Fix something.

## 0.26.2

* Previous fix.
"""

    body = get_release_notes_body(content, "0.26.3", Path("docs/release-notes.md"))

    assert body == "## Highlights\n\n* Fix something.\n"


def test_get_release_notes_body_requires_version_section() -> None:
    content = "# Release Notes\n\n## Latest Changes\n"

    with pytest.raises(RuntimeError, match="Could not find"):
        get_release_notes_body(content, "0.26.3", Path("docs/release-notes.md"))


def test_get_release_notes_body_requires_non_empty_section() -> None:
    content = """# Release Notes

## Latest Changes

## 0.26.3

## 0.26.2

* Previous fix.
"""

    with pytest.raises(RuntimeError, match="is empty"):
        get_release_notes_body(content, "0.26.3", Path("docs/release-notes.md"))


def test_cli_updates_configured_files(tmp_path: Path) -> None:
    version_file = tmp_path / "package" / "__init__.py"
    version_file.parent.mkdir()
    version_file.write_text('__version__ = "0.26.2"\n')
    release_notes_file = tmp_path / "release-notes.md"
    release_notes_file.write_text(
        """# Release Notes

## Latest Changes

### Fixes

* Fix something.
"""
    )

    result = runner.invoke(
        app,
        [
            "prepare",
            "patch",
            "--version-file",
            str(version_file),
            "--release-notes-file",
            str(release_notes_file),
            "--date",
            "2026-05-28",
        ],
    )

    assert result.exit_code == 0, result.output
    assert "Prepared release 0.26.3 (2026-05-28)" in result.output
    assert version_file.read_text() == '__version__ = "0.26.3"\n'
    assert "## 0.26.3 (2026-05-28)" in release_notes_file.read_text()


def test_cli_accepts_env_vars(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    version_file = tmp_path / "package" / "__init__.py"
    version_file.parent.mkdir()
    version_file.write_text('__version__ = "0.26.2"\n')
    release_notes_file = tmp_path / "docs" / "release-notes.md"
    release_notes_file.parent.mkdir()
    release_notes_file.write_text("# Release Notes\n\n## Latest Changes\n")
    monkeypatch.setenv("PREPARE_RELEASE_BUMP", "minor")
    monkeypatch.setenv("PREPARE_RELEASE_VERSION_FILE", str(version_file))
    monkeypatch.setenv("PREPARE_RELEASE_RELEASE_NOTES_FILE", str(release_notes_file))
    monkeypatch.setenv("PREPARE_RELEASE_DATE", "2026-05-28")

    result = runner.invoke(app, ["prepare"])

    assert result.exit_code == 0, result.output
    assert "Prepared release 0.27.0 (2026-05-28)" in result.output
    assert version_file.read_text() == '__version__ = "0.27.0"\n'
    assert "## 0.27.0 (2026-05-28)" in release_notes_file.read_text()


def test_cli_prints_current_version(tmp_path: Path) -> None:
    version_file = tmp_path / "package" / "__init__.py"
    version_file.parent.mkdir()
    version_file.write_text('__version__ = "0.26.2"\n')

    result = runner.invoke(
        app,
        [
            "current-version",
            "--version-file",
            str(version_file),
        ],
    )

    assert result.exit_code == 0, result.output
    assert result.output == "0.26.2\n"


def test_cli_prints_release_notes(tmp_path: Path) -> None:
    version_file = tmp_path / "package" / "__init__.py"
    version_file.parent.mkdir()
    version_file.write_text('__version__ = "0.26.3"\n')
    release_notes_file = tmp_path / "release-notes.md"
    release_notes_file.write_text(
        """# Release Notes

## Latest Changes

## 0.26.3 (2026-05-28)

### Fixes

* Fix something.
"""
    )

    result = runner.invoke(
        app,
        [
            "release-notes",
            "--version-file",
            str(version_file),
            "--release-notes-file",
            str(release_notes_file),
        ],
    )

    assert result.exit_code == 0, result.output
    assert result.output == "### Fixes\n\n* Fix something.\n"



"""
Tests for the Progress bar functionality.
Created after vendoring Click to ensure test coverage is back up to 100%.
"""

import io
import shutil

import pytest
import typer
from typer import progressbar
from typer._click import _termui_impl
from typer.testing import CliRunner

runner = CliRunner()


def _fake_clock(monkeypatch: pytest.MonkeyPatch) -> list[float]:
    clock = [0.0]
    monkeypatch.setattr(_termui_impl.time, "time", lambda: clock[0])
    return clock


def _pbar(**kw):
    return progressbar(file=kw.pop("file", io.StringIO()), **kw)


@pytest.mark.parametrize(
    ("iterable", "length", "hidden", "label", "expected_count"),
    [
        (["a", "b"], None, False, "Processing", 2),
        (None, 3, False, "Counting", 3),
        (["x", "y"], None, True, "Hidden", 2),
    ],
)
def test_progressbar(iterable, length, hidden, label, expected_count):
    app = typer.Typer()

    @app.command()
    def main():
        bar_out = io.StringIO()
        count = 0
        with progressbar(
            iterable, length=length, hidden=hidden, label=label, file=bar_out
        ) as bar:
            for _ in bar:
                count += 1
        typer.echo(f"count={count}")
        typer.echo(f"bar={bar_out.getvalue()!r}")

    result = runner.invoke(app, [])
    assert result.exit_code == 0, result.output
    assert f"count={expected_count}" in result.stdout
    assert (label in result.stdout) == (not hidden)


@pytest.mark.parametrize(
    ("label", "pbar_kw", "must_contain", "must_not_contain"),
    [
        pytest.param(
            "TTY",
            {
                "show_pos": True,
                "show_percent": True,
                "item_show_func": lambda item: f"item={item}",
            },
            ("TTY", "1/1", "100%", "item=x"),
            (),
        ),
        pytest.param(
            "HeurPct",
            {},
            ("HeurPct", "100%"),
            ("1/1",),
        ),
        pytest.param(
            "HeurPos",
            {"show_pos": True},
            ("HeurPos", "1/1"),
            ("100%",),
        ),
    ],
)
def test_progressbar_tty(
    monkeypatch, label: str, pbar_kw: dict, must_contain, must_not_contain
):
    monkeypatch.setattr(_termui_impl, "isatty", lambda f: True)
    _fake_clock(monkeypatch)

    app = typer.Typer()

    @app.command()
    def main():
        bar_out = io.StringIO()
        with progressbar(
            ["x"],
            label=label,
            file=bar_out,
            bar_template="%(label)s %(info)s",
            width=1,
            **pbar_kw,
        ) as bar:
            for _ in bar:
                pass
        typer.echo(bar_out.getvalue())

    result = runner.invoke(app, [])
    assert result.exit_code == 0, result.output
    for part in must_contain:
        assert part in result.stdout
    for part in must_not_contain:
        assert part not in result.stdout


def test_progressbar_tty_show_eta(monkeypatch):
    monkeypatch.setattr(_termui_impl, "isatty", lambda f: True)
    clock = _fake_clock(monkeypatch)
    clock[0] = 1_000.0

    app = typer.Typer()

    @app.command()
    def main():
        bar_out = io.StringIO()
        with progressbar(
            ["a", "b"],
            label="ETA",
            file=bar_out,
            show_pos=True,
            show_percent=False,
            show_eta=True,
            bar_template="%(label)s %(info)s",
            width=1,
        ) as bar:
            for i, _ in enumerate(bar):
                if i == 0:
                    clock[0] = 1_001.0
        typer.echo(bar_out.getvalue())

    result = runner.invoke(app, [])
    assert result.exit_code == 0, result.output
    for part in ("ETA", "1/2", "00:00:01"):
        assert part in result.stdout


def test_progressbar_autowidth(monkeypatch):
    monkeypatch.setattr(_termui_impl, "isatty", lambda f: True)
    call = [0]
    real_get_terminal_size = shutil.get_terminal_size

    def fake_get_terminal_size(*args, **kwargs):
        # Pytest (and others) call get_terminal_size(fallback=...); only stub no-arg calls
        if args or kwargs:
            return real_get_terminal_size(*args, **kwargs)
        col = 120 if call[0] == 0 else 40
        call[0] += 1
        return type("TS", (), {"columns": col, "lines": 24})()

    monkeypatch.setattr(shutil, "get_terminal_size", fake_get_terminal_size)

    state: dict[str, object] = {}

    app = typer.Typer()

    @app.command()
    def main():
        out = io.StringIO()
        with progressbar(["a", "b"], width=0, label="AW", file=out) as bar:
            state["autowidth"] = bar.autowidth
            for _ in bar:
                pass
        state["call_count"] = call[0]
        state["out"] = out.getvalue()
        typer.echo("done")

    result = runner.invoke(app, [])
    assert result.exit_code == 0, result.output
    assert state["autowidth"] is True
    assert state["call_count"] >= 2
    out = str(state["out"])
    assert "\r" in out and "AW" in out
    assert "0%" in out and "50%" in out and "100%" in out


def test_progress_bar_iter():
    not_entered = _pbar(iterable=[1, 2], length=2)
    with pytest.raises(RuntimeError, match="with block"):
        iter(not_entered)

    entered = _pbar(iterable=[10, 20], length=2)
    with entered:
        iterator = iter(entered)
        assert next(iterator) == 10
        assert next(entered) == 20
        with pytest.raises(StopIteration):
            next(iterator)


def test_progress_bar_time(monkeypatch):
    clock = _fake_clock(monkeypatch)
    state: dict[str, object] = {}
    clock[0] = 1_000.0

    app = typer.Typer()

    @app.command()
    def main():
        bar = _pbar(iterable=None, length=10)
        state["tpi0"] = bar.time_per_iteration
        clock[0] = 1_000.5
        bar.make_step(1)
        state["avg_after_one"] = list(bar.avg)
        state["tpi_after_one"] = bar.time_per_iteration
        clock[0] = 1_001.0
        bar.make_step(1)
        state["pos2"] = bar.pos
        state["avg2"] = list(bar.avg)
        state["tpi2"] = bar.time_per_iteration
        clock[0] = 1_002.0
        bar.make_step(1)
        state["avg3"] = list(bar.avg)
        state["tpi3"] = bar.time_per_iteration
        typer.echo("ok")

    result = runner.invoke(app, [])
    assert result.exit_code == 0, result.output
    assert state["tpi0"] == 0.0
    assert state["avg_after_one"] == [] and state["tpi_after_one"] == 0.0
    assert state["pos2"] == 2
    assert state["avg2"] == [(1_001.0 - 1_000.0) / 2.0]
    assert state["tpi2"] == pytest.approx(0.5)
    assert state["avg3"] == [0.5, (1_002.0 - 1_000.0) / 3.0]
    assert state["tpi3"] == pytest.approx(sum(state["avg3"]) / 2.0)  # type: ignore[arg-type]


def test_progress_bar_time_zero_steps(monkeypatch):
    clock = _fake_clock(monkeypatch)
    state: dict[str, object] = {}
    clock[0] = 2_000.0

    app = typer.Typer()

    @app.command()
    def main():
        bar = _pbar(iterable=None, length=3)
        clock[0] = 2_001.0
        bar.make_step(0)
        state["pos"] = bar.pos
        state["avg"] = list(bar.avg)
        state["tpi"] = bar.time_per_iteration
        typer.echo("ok")

    result = runner.invoke(app, [])
    assert result.exit_code == 0, result.output
    assert state["pos"] == 0
    assert state["avg"] == [1.0]
    assert state["tpi"] == pytest.approx(1.0)


def test_progress_bar_eta(monkeypatch):
    state: dict[str, object] = {}

    app = typer.Typer()

    @app.command()
    def main():
        state["eta0"] = _pbar(iterable=[1, 2], length=None).eta

        done = _pbar(iterable=None, length=5)
        done.pos, done.finished = 2, True
        state["eta_done"] = done.eta

        fresh = _pbar(iterable=None, length=5)
        state["eta_known_fresh"] = fresh.eta_known
        state["fmt_eta_fresh"] = fresh.format_eta()

        clock = _fake_clock(monkeypatch)
        clock[0] = 5_000.0
        bar = _pbar(iterable=None, length=10)
        clock[0] = 5_001.0
        bar.make_step(3)
        state["pos3"] = bar.pos
        state["eta_after"] = bar.eta
        state["tpi"] = bar.time_per_iteration

        cases_out = []
        for t0, t1, length, n_steps, _expected_fmt, expected_eta_int in (
            (9_000.0, 9_001.0, 10, 1, "00:00:09", None),
            (1_000.0, 100_000.0, 2, 1, "1d 03:30:00", 99_000),
        ):
            clock2 = _fake_clock(monkeypatch)
            clock2[0] = t0
            b = _pbar(iterable=None, length=length)
            clock2[0] = t1
            b.make_step(n_steps)
            cases_out.append(
                (
                    b.eta_known,
                    b.format_eta(),
                    int(b.eta) if expected_eta_int is not None else None,
                )
            )

        state["cases"] = cases_out

        clock3 = _fake_clock(monkeypatch)
        clock3[0] = 3_000.0
        b2 = _pbar(iterable=None, length=2)
        clock3[0] = 3_001.0
        b2.make_step(1)
        state["fmt_before_finish"] = b2.format_eta()
        b2.finish()
        state["fmt_after_finish"] = b2.format_eta()
        typer.echo("ok")

    result = runner.invoke(app, [])
    assert result.exit_code == 0, result.output
    assert state["eta0"] == 0.0
    assert state["eta_done"] == 0.0
    assert not state["eta_known_fresh"] and state["fmt_eta_fresh"] == ""
    assert state["pos3"] == 3
    assert state["eta_after"] == pytest.approx(state["tpi"] * (10 - 3))  # type: ignore[operator]

    (ek1, fmt1, ei1), (ek2, fmt2, ei2) = state["cases"]  # type: ignore[misc]
    assert ek1 and fmt1 == "00:00:09" and ei1 is None
    assert ek2 and fmt2 == "1d 03:30:00" and ei2 == 99_000

    assert state["fmt_before_finish"] != ""
    assert state["fmt_after_finish"] == ""


@pytest.mark.parametrize(
    ("width", "fill_char", "empty_char", "expected_bar", "finished", "sample_timing"),
    [
        pytest.param(4, "X", "-", "XXXX", True, False, id="finished"),
        pytest.param(4, "#", "-", "----", False, False, id="no_timing_yet"),
        pytest.param(5, "*", ".", None, False, True, id="indeterminate"),
    ],
)
def test_progress_bar_unknown_length(
    monkeypatch,
    width: int,
    fill_char: str,
    empty_char: str,
    expected_bar: str | None,
    finished: bool,
    sample_timing: bool,
):
    clock: list[float] | None = _fake_clock(monkeypatch) if sample_timing else None
    if clock is not None:
        clock[0] = 100.0

    state: dict[str, object] = {}

    class _IterableWithoutLength:
        def __iter__(self):
            return iter((1, 2, 3))

    app = typer.Typer()

    @app.command()
    def main():
        bar = _pbar(
            iterable=_IterableWithoutLength(),
            length=None,
            width=width,
            fill_char=fill_char,
            empty_char=empty_char,
        )
        assert bar.length is None

        if sample_timing:
            assert clock is not None
            clock[0] = 101.0
            bar.make_step(1)
            assert bar.time_per_iteration > 0
            rendered = bar.format_bar()
            assert len(rendered) == width
            assert rendered.count(fill_char) == 1
            assert rendered.count(empty_char) == width - 1
            state["branch"] = "sample_timing"
        elif finished:
            bar.finished = True
            state["bar"] = bar.format_bar()
            state["branch"] = "finished"
        else:
            assert bar.time_per_iteration == 0.0
            state["bar"] = bar.format_bar()
            state["branch"] = "no_timing"
        typer.echo("ok")

    result = runner.invoke(app, [])
    assert result.exit_code == 0, result.output
    if sample_timing:
        assert state["branch"] == "sample_timing"
    else:
        assert state["bar"] == expected_bar



import subprocess
import sys

import typer
from typer.testing import CliRunner

runner = CliRunner()


def test_script_help():
    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "coverage",
            "run",
            "-m",
            "typer",
            "tests/assets/cli/sample.py",
            "--help",
        ],
        capture_output=True,
        encoding="utf-8",
    )
    assert "run" in result.stdout


def test_not_python():
    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "coverage",
            "run",
            "-m",
            "typer",
            "tests/assets/cli/not_python.txt",
            "run",
        ],
        capture_output=True,
        encoding="utf-8",
    )
    assert "Could not import as Python file" in result.stderr


def test_short_help() -> None:
    app = typer.Typer(
        rich_markup_mode=None,
        context_settings={"max_content_width": 50},
    )

    @app.command(help=" \n\t ")
    def empty() -> None:
        pass  # pragma: no cover

    @app.command(help="\b first sentence.")
    def marker() -> None:
        pass  # pragma: no cover

    # Forcing truncation
    @app.command(help=f"{'x' * 30} {'y' * 5} z trailing")
    def long() -> None:
        pass  # pragma: no cover

    result = runner.invoke(app, ["--help"], terminal_width=50)
    assert result.exit_code == 0
    assert "empty" in result.output
    assert "marker" in result.output
    assert "long" in result.output
    assert "first sentence." in result.output
    assert f"{'x' * 30}..." in result.output


def test_help_wrapping() -> None:
    app = typer.Typer(
        rich_markup_mode=None,
        context_settings={"max_content_width": 50},
    )

    @app.command(
        help=(
            "Wrapped paragraph has enough words to wrap in help output.\n"
            "\n"
            "\n"
            "\b\n"
            "RAW-LINE-ONE stays on one line even with many many many words.\n"
            "RAW-LINE-TWO keeps original formatting.\n"
            "\n"
            "Final paragraph wraps normally as well."
        )
    )
    def cmd() -> None:
        pass  # pragma: no cover

    result = runner.invoke(app, ["cmd", "--help"], terminal_width=50)
    assert result.exit_code == 0
    assert "Wrapped paragraph has enough words to wrap" in result.output
    assert (
        "RAW-LINE-ONE stays on one line even with many many many words."
        in result.output
    )
    assert "RAW-LINE-TWO keeps original formatting." in result.output
    assert "Final paragraph wraps normally as well." in result.output


def test_help_wrapping_long_name() -> None:
    app = typer.Typer(rich_markup_mode=None)

    @app.command()
    def cmd(value: str) -> None:
        pass  # pragma: no cover

    result = runner.invoke(
        app,
        ["cmd", "--help"],
        terminal_width=40,
        prog_name="very-long-program-name-that-forces-wrap",
    )
    assert result.exit_code == 0

    output_lines = result.output.splitlines()
    usage_idx = output_lines.index("Usage: very-long-program-name-that-forces-wrap ")
    args_line = output_lines[usage_idx + 1]
    assert args_line.lstrip() == "[OPTIONS] {value}"
    assert args_line.startswith(" ")


def test_format_long_help_option() -> None:
    app = typer.Typer(rich_markup_mode=None)

    @app.command()
    def cmd(
        very_long: str = typer.Option(
            ...,
            "--this-is-a-very-very-very-long-option-name",
            help="Description is rendered in the next line for long option labels.",
        ),
    ) -> None:
        pass  # pragma: no cover

    result = runner.invoke(app, ["cmd", "--help"], terminal_width=80)
    assert result.exit_code == 0

    output_lines = result.output.splitlines()
    option_idx = next(
        i
        for i, line in enumerate(output_lines)
        if "--this-is-a-very-very-very-long-option-name" in line
    )
    assert "Description is rendered" not in output_lines[option_idx]
    first_desc_line = output_lines[option_idx + 1]
    assert first_desc_line.lstrip().startswith("Description is rendered")
    continuation_block = " ".join(
        line.strip() for line in output_lines[option_idx + 1 :] if line.startswith(" ")
    )
    assert (
        "Description is rendered in the next line for long option labels."
        in continuation_block
    )


def test_typer_run_usage() -> None:
    from typer.cli import app as typer_cli_app

    result = runner.invoke(
        typer_cli_app,
        ["tests/assets/cli/minimum_main.py", "run"],
        prog_name="typer",
    )

    assert result.exit_code == 2
    usage_line = result.output.splitlines()[0]
    assert usage_line.startswith("Usage: typer [PATH_OR_MODULE] run [OPTIONS] {name}")
    assert "Try 'typer [PATH_OR_MODULE] run --help' for help." in result.output



import subprocess
import sys
from io import BytesIO, StringIO
from pathlib import Path

import pytest
import typer
from typer._click._compat import get_best_encoding, should_strip_ansi
from typer._click.utils import PacifyFlushWrapper
from typer.testing import CliRunner

from tests.utils import needs_linux, needs_windows

app = typer.Typer()


@app.command()
def read_text(file_in: typer.FileText = typer.Option(..., lazy=True)):
    data = file_in.read()
    typer.echo(f"text-len={len(data)}")


@app.command()
def write_text(file_out: typer.FileTextWrite = typer.Option(..., lazy=None)):
    file_out.write("This is a single line\n")
    typer.echo("1 line written")


@app.command()
def write_lazy(file_out: typer.FileTextWrite = typer.Option(..., lazy=True)):
    file_out.write("This is a single lazy line\n")
    typer.echo("1 line written")


@app.command()
def probe_lazy_file_behaviors(
    file_in: typer.FileText = typer.Option(..., lazy=True),
    file_out: typer.FileTextWrite = typer.Option(..., lazy=True),
):
    typer.echo(f"repr-before={repr(file_out)}")
    file_out.write("repr-opened\n")
    typer.echo(f"repr-after={repr(file_out)}")
    with file_in as stream:
        typer.echo(f"context-len={len(stream.read())}")
        stream.seek(0)
        first_line = next(iter(stream), "")
        typer.echo(f"first-line={first_line.rstrip()}")


@app.command()
def write_binary(file_out: typer.FileBinaryWrite = typer.Option(...)):
    file_out.write(b"binary-written\n")


@app.command()
def write_binary_stderr():
    stream = typer.get_binary_stream("stderr")
    stream.write(b"binary-stderr\n")
    stream.flush()


@app.command()
def read_binary(file_in: typer.FileBinaryRead = typer.Option(...)):
    data = file_in.read()
    typer.echo(f"binary-len={len(data)}")


runner = CliRunner()


def test_text_stdin_dash() -> None:
    result = runner.invoke(app, ["read-text", "--file-in=-"], input="hello\n")
    assert result.exit_code == 0
    assert "text-len=6" in result.output


def test_lazy_file(tmp_path: Path) -> None:
    # dash: written to stdout
    result = runner.invoke(app, ["write-text", "--file-out=-"])
    assert result.exit_code == 0
    assert "This is a single line" in result.output
    assert "1 line written" in result.output

    # lazy + file
    file_path = tmp_path / "example.txt"
    result = runner.invoke(app, ["write-lazy", f"--file-out={file_path}"])
    assert result.exit_code == 0
    assert "This is a single lazy line" not in result.output
    assert "1 line written" in result.output
    assert file_path.exists()
    assert file_path.read_text() == "This is a single lazy line\n"

    # lazy probe: unopened/opened repr, context manager, and iteration.
    result = runner.invoke(
        app,
        [
            "probe-lazy-file-behaviors",
            f"--file-in={file_path}",
            f"--file-out={tmp_path / 'repr-opened.txt'}",
        ],
    )
    assert result.exit_code == 0
    assert "repr-before=<unopened file" in result.output
    assert "repr-after=<unopened file" not in result.output
    assert "repr-opened.txt" in result.output
    assert "context-len=27" in result.output
    assert "first-line=This is a single lazy line" in result.output

    # lazy + dash: written to stdout (lazy setting is pretty much ignored)
    result = runner.invoke(app, ["write-lazy", "--file-out=-"])
    assert result.exit_code == 0
    assert "This is a single lazy line" in result.output
    assert "1 line written" in result.output


def test_filelike_conversion() -> None:
    stream = StringIO()
    result = runner.invoke(
        app, ["write-text"], default_map={"write-text": {"file_out": stream}}
    )
    assert result.exit_code == 0
    assert "1 line written" in result.output
    assert stream.getvalue() == "This is a single line\n"


def test_binary_dash() -> None:
    result = runner.invoke(app, ["write-binary", "--file-out=-"])
    assert result.exit_code == 0
    assert result.stdout_bytes == b"binary-written\n"

    result = runner.invoke(
        app, ["read-binary", "--file-in=-"], input=b"\x00\x01\x02abc"
    )
    assert result.exit_code == 0
    assert "binary-len=6" in result.output


def test_binary_stderr() -> None:
    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "coverage",
            "run",
            __file__,
            "write-binary-stderr",
        ],
        capture_output=True,
    )
    assert result.returncode == 0
    assert result.stderr == b"binary-stderr\n"


@pytest.mark.parametrize(
    ("errors_arg", "expected_errors"),
    [
        (None, "replace"),
        ("strict", "strict"),
    ],
)
def test_get_text_stream_errors(
    monkeypatch,
    errors_arg: str | None,
    expected_errors: str,
) -> None:
    class BinaryStdout(BytesIO):
        pass

    binary_stdout = BinaryStdout()
    monkeypatch.setattr(sys, "stdout", binary_stdout)

    text_stream = typer.get_text_stream("stdout", encoding=None, errors=errors_arg)
    text_stream.write("stream-text")
    text_stream.flush()

    assert text_stream.errors == expected_errors
    assert text_stream.writable() is True
    assert binary_stdout.getvalue() == b"stream-text"


def test_get_best_encoding() -> None:
    """Test that ASCII is being transformed into UTF-8"""

    class AsciiStream:
        encoding = "ascii"

    class Utf8Stream:
        encoding = "utf-8"

    class UnknownStream:
        encoding = "unknown"

    assert get_best_encoding(AsciiStream()) == "utf-8"
    assert get_best_encoding(Utf8Stream()) == "utf-8"
    assert get_best_encoding(UnknownStream()) == "unknown"


def test_pacify_flush_wrapper() -> None:
    class Wrapped:
        def __init__(self) -> None:
            self.name = "wrapped-stream"

        def flush(self) -> None:
            return None  # pragma: no cover

    wrapped = PacifyFlushWrapper(Wrapped())
    assert wrapped.name == "wrapped-stream"


def test_text_stream_isatty(monkeypatch) -> None:
    class BinaryStdout(BytesIO):
        def isatty(self) -> bool:
            return True

    binary_stdout = BinaryStdout()
    monkeypatch.setattr(sys, "stdout", binary_stdout)
    text_stream = typer.get_text_stream("stdout", encoding="utf-8", errors=None)
    assert text_stream.isatty() is True


def test_text_stream_buffer_read1(monkeypatch) -> None:
    class BinaryStdinNoRead1:
        def __init__(self, data: bytes) -> None:
            self._data = data
            self._pos = 0

        def read(self, size: int = -1) -> bytes:
            if size < 0:
                size = len(self._data) - self._pos  # pragma: no cover
            chunk = self._data[self._pos : self._pos + size]
            self._pos += len(chunk)
            return chunk

    binary_stdin = BinaryStdinNoRead1(b"hello")
    monkeypatch.setattr(sys, "stdin", binary_stdin)
    text_stream = typer.get_text_stream("stdin", encoding="utf-8", errors=None)
    assert text_stream._stream.read1(4) == b"hell"


def test_binary_stream(monkeypatch) -> None:
    binary_stdin = BytesIO(b"hello")
    binary_stdout = BytesIO()
    monkeypatch.setattr(sys, "stdin", binary_stdin)
    monkeypatch.setattr(sys, "stdout", binary_stdout)

    assert typer.get_binary_stream("stdin") is binary_stdin
    assert typer.get_binary_stream("stdout") is binary_stdout


def test_binary_stream_raises(monkeypatch) -> None:
    class TextOnlyStdin:
        def read(self, n: int = -1) -> str:
            return "hello"

    monkeypatch.setattr(sys, "stdin", TextOnlyStdin())
    with pytest.raises(RuntimeError, match="Was not able to determine binary stream"):
        typer.get_binary_stream("stdin")


def test_stream_unknown() -> None:
    with pytest.raises(TypeError, match="Unknown standard stream 'Plumbus'"):
        typer.get_binary_stream("Plumbus")  # type: ignore[arg-type]

    with pytest.raises(TypeError, match="Unknown standard stream 'Fleeb'"):
        typer.get_text_stream("Fleeb")  # type: ignore[arg-type]


def test_format_filename() -> None:
    filename = b"folder/subdir/demo.txt"
    assert typer.format_filename(filename, shorten=True) == "demo.txt"


def test_file_error(monkeypatch, tmp_path: Path) -> None:
    file_path = tmp_path / "cannot-open.txt"

    def fake_open(path, *args, **kwargs):
        if Path(path) == file_path:
            raise OSError()

    monkeypatch.setattr("builtins.open", fake_open)
    result = runner.invoke(app, ["write-text", f"--file-out={file_path}"])
    assert result.exit_code == 1
    assert "Could not open file" in result.output
    assert "cannot-open.txt" in result.output
    assert "unknown error" in result.output


@needs_windows
def test_app_dir_windows_fallback(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.delenv("APPDATA", raising=False)
    monkeypatch.setattr("os.path.expanduser", lambda _path: r"C:\Users\Tester")

    assert typer.get_app_dir("My App", roaming=True) == r"C:\Users\Tester\My App"


@needs_linux
def test_app_dir_force_posix(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("os.path.expanduser", lambda _path: "/home/tester/.my-app")

    assert typer.get_app_dir("My App", force_posix=True) == "/home/tester/.my-app"


def test_text_stream_binary_buffer(monkeypatch) -> None:
    class TextStdinWithBinaryBuffer:
        def __init__(self, data: bytes) -> None:
            self.buffer = BytesIO(data)
            self.encoding = "latin-1"

        def read(self, n: int = -1) -> str:
            raise OSError("text stream is not readable directly")

    class TextStdoutWithBinaryBuffer:
        def __init__(self) -> None:
            self.buffer = BytesIO()
            self.encoding = "latin-1"

        def write(self, s: str) -> int:
            raise OSError("text stream is not writable directly")

    stdin = TextStdinWithBinaryBuffer(b"hello")
    stdout = TextStdoutWithBinaryBuffer()

    monkeypatch.setattr(sys, "stdin", stdin)
    monkeypatch.setattr(sys, "stdout", stdout)

    text_stdin = typer.get_text_stream("stdin", encoding="utf-8", errors=None)
    text_stdout = typer.get_text_stream("stdout", encoding="utf-8", errors=None)

    assert text_stdin.read() == "hello"
    text_stdout.write("ok")
    text_stdout.flush()
    assert stdout.buffer.getvalue() == b"ok"


def test_text_stream_binary_stream(monkeypatch) -> None:
    binary_stdout = BytesIO()
    monkeypatch.setattr(sys, "stdout", binary_stdout)
    text_stream = typer.get_text_stream("stdout", encoding="utf-8", errors=None)
    text_stream.write("ok")
    text_stream.flush()
    assert binary_stdout.getvalue() == b"ok"


def test_text_stream_stdout_no_binary(
    monkeypatch,
) -> None:
    class TextStdoutNoBinaryFallback:
        encoding = "utf-8"
        errors = "strict"

        def write(self, s: str) -> int:
            if isinstance(s, bytes):
                raise TypeError("bytes not supported")
            return len(s)

    stdout = TextStdoutNoBinaryFallback()
    monkeypatch.setattr(sys, "stdout", stdout)
    text_stream = typer.get_text_stream("stdout", encoding="utf-8", errors="replace")
    assert text_stream is stdout


def test_jupyter_wrapped_stream(monkeypatch) -> None:
    class JupyterLikeStdout(BytesIO):
        __module__ = "ipykernel.iostream"

        def isatty(self) -> bool:
            return False

    binary_stdout = JupyterLikeStdout()
    monkeypatch.setattr(sys, "stdout", binary_stdout)
    text_stream = typer.get_text_stream("stdout", encoding="utf-8", errors=None)
    assert should_strip_ansi(text_stream, color=None) is False


def test_should_strip_ansi(monkeypatch) -> None:
    class NonTtyStdin(BytesIO):
        def isatty(self) -> bool:
            return False

    stdin = NonTtyStdin()
    monkeypatch.setattr(sys, "stdin", stdin)
    assert should_strip_ansi(stream=None, color=None) is True
    assert should_strip_ansi(stream=None, color=True) is False
    assert should_strip_ansi(stream=None, color=False) is True


if __name__ == "__main__":
    app()



from typing import Annotated

import pytest
import typer
import typer._completion_shared
import typer.completion
from typer import _click
from typer.core import TyperArgument, TyperCommand, TyperGroup, TyperOption, _split_opt
from typer.testing import CliRunner

runner = CliRunner()


def test_human_readable_name() -> None:
    app = typer.Typer()

    @app.command()
    def main(
        my_arg_1: Annotated[str, typer.Argument()],
        my_arg_2: Annotated[str, typer.Argument(metavar="META_ARG")],
        my_opt: Annotated[str, typer.Option()],
    ):
        pass  # pragma: no cover

    command = typer.main.get_command(app)
    params = {param.name: param for param in command.params}

    assert params["my_arg_1"].human_readable_name == "my_arg_1"
    assert params["my_arg_2"].human_readable_name == "META_ARG"
    assert params["my_opt"].human_readable_name == "my_opt"


def test_parameter_metavar() -> None:
    app = typer.Typer(rich_markup_mode=None)

    @app.command()
    def cmd(name: Annotated[str, typer.Option(metavar="CUSTOM")]) -> None:
        pass  # pragma: no cover

    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "--name CUSTOM" in result.output


def test_parameter_nargs_gt_1() -> None:
    param = TyperArgument(param_decls=["value"], type=str, nargs=2)
    ctx = _click.Context(TyperCommand(name="cmd"))

    assert param.type_cast_value(ctx, ("one", "two")) == ("one", "two")

    with pytest.raises(
        _click.exceptions.BadParameter, match="Takes 2 values but 1 given."
    ):
        param.type_cast_value(ctx, ("one",))


def test_parameter_constructor() -> None:
    # no param_decl and expose_value is False: sets name to None
    arg = TyperArgument(param_decls=[], expose_value=False)
    assert arg.name is None
    assert arg.opts == []
    assert arg.secondary_opts == []

    # no param_decl and expose_value is True: raises
    with pytest.raises(TypeError, match="does not have a name."):
        TyperArgument(param_decls=[], expose_value=True)

    # len(param_decl) > 1: raises
    with pytest.raises(TypeError, match="take exactly one parameter declaration"):
        TyperArgument(param_decls=["first", "second"])

    # duplicated identifier in option declarations: raises
    with pytest.raises(TypeError, match="Name 'name' defined twice"):
        TyperOption(param_decls=["name", "name"], required=False)

    # same true/false flag in boolean option declaration: raises
    with pytest.raises(ValueError, match="cannot use the same flag for true/false"):
        TyperOption(param_decls=["flag", "--flag/--flag"], required=False, is_flag=True)

    # inferred name is not a valid identifier: sets name to None
    unnamed_option = TyperOption(param_decls=["--123"], required=False)
    assert unnamed_option.name is None

    # no param_decl and prompt=True: raises
    with pytest.raises(TypeError, match="'name' is required with 'prompt=True'."):
        TyperOption(param_decls=[], expose_value=False, prompt=True, required=False)

    # count works
    option = TyperOption(
        param_decls=["verbose", "--verbose", "-v"],
        type=None,
        default=0,
        required=False,
        count=True,
    )
    assert isinstance(option.type, _click.types.IntRange)
    assert option.type.min == 0


def test_option_error_hint() -> None:
    option = TyperOption(
        param_decls=["name", "--name"],
        required=False,
        show_envvar=True,
        envvar="APP_NAME",
    )
    hint = option.get_error_hint(_click.Context(TyperCommand(name="cmd")))
    assert "(env var: 'APP_NAME')" in hint


def test_group_init() -> None:
    group_no_commands = TyperGroup(name="root", commands=None)
    assert group_no_commands.commands == {}

    named = TyperCommand(name="named")
    unnamed = TyperCommand(name=None)
    group_command_sequence = TyperGroup(name="root", commands=[named, unnamed])
    assert group_command_sequence.commands == {"named": named}


@pytest.mark.parametrize("with_result_callback", [False, True])
def test_group_result_callback(with_result_callback: bool) -> None:
    called = {"child": False, "result_callback": False}

    def child_callback() -> None:
        called["child"] = True
        return None

    def result_callback(value, **kwargs):  # type: ignore[no-untyped-def]
        called["result_callback"] = True
        return value

    child = TyperCommand(name="child", callback=child_callback)
    group = TyperGroup(
        name="root",
        commands={"child": child},
        result_callback=result_callback if with_result_callback else None,
    )
    ctx = group.make_context("root", ["child"])

    result = group.invoke(ctx)

    assert result is None
    assert called["child"] is True
    assert called["result_callback"] is with_result_callback
    assert ctx.invoked_subcommand == "child"


def test_group_add_command() -> None:
    group = TyperGroup(name="root")
    unnamed_command = TyperCommand(name=None)

    with pytest.raises(TypeError, match="Command has no name."):
        group.add_command(unnamed_command)


def test_group_click_resolve_command() -> None:
    child = TyperCommand(name="child")
    group = TyperGroup(name="root", commands={"child": child})
    ctx = group.make_context("root", ["CHILD"], token_normalize_func=str.lower)

    cmd_name, cmd, remaining = group._click_resolve_command(ctx, ["CHILD"])

    assert cmd_name == "child"
    assert cmd is child
    assert remaining == []


@pytest.mark.parametrize(
    ("envvar", "auto_prefix", "set_env", "expected"),
    [
        ("APP_NAME", None, True, "my-precious"),
        (None, "APP", True, "my-precious"),
        (None, None, False, None),
    ],
)
def test_option_resolve_envvar(
    monkeypatch: pytest.MonkeyPatch,
    envvar: str | None,
    auto_prefix: str | None,
    set_env: bool,
    expected: str | None,
) -> None:
    option = TyperOption(
        param_decls=["name", "--name"],
        required=False,
        envvar=envvar,
    )
    if set_env:
        monkeypatch.setenv("APP_NAME", "my-precious")

    ctx = _click.Context(TyperCommand(name="cmd"), auto_envvar_prefix=auto_prefix)
    assert option.resolve_envvar_value(ctx) == expected


def test_option_resolve_envvar_list(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    option = TyperOption(
        param_decls=["name", "--name"],
        required=False,
        envvar=["APP_NAME_1", "APP_NAME_2"],
    )
    monkeypatch.delenv("APP_NAME_1", raising=False)
    monkeypatch.delenv("APP_NAME_2", raising=False)
    ctx = _click.Context(TyperCommand(name="cmd"))

    assert option.resolve_envvar_value(ctx) is None


def test_context_auto_envvar() -> None:
    app = typer.Typer(context_settings={"auto_envvar_prefix": "APP"})
    sub_app = typer.Typer()

    @sub_app.command()
    def clone(ctx: typer.Context) -> None:
        print(ctx.auto_envvar_prefix)

    app.add_typer(sub_app, name="beth")

    result = runner.invoke(app, ["beth", "clone"])
    assert result.exit_code == 0
    assert "APP_BETH_CLONE" in result.stdout


def test_context_with_resource() -> None:
    events: list[str] = []

    class DemoResource:
        def __enter__(self) -> str:
            events.append("enter")
            return "pickle-rick"

        def __exit__(self, *args: object) -> None:
            events.append("exit")

    app = typer.Typer()

    @app.command()
    def cmd(ctx: typer.Context) -> None:
        value = ctx.with_resource(DemoResource())
        assert value == "pickle-rick"
        assert events == ["enter"]
        print("I'm a pickle")

    result = runner.invoke(app)

    assert result.exit_code == 0
    assert "I'm a pickle" in result.stdout
    assert events == ["enter", "exit"]


def test_context_find_root() -> None:
    app = typer.Typer()
    sub_app = typer.Typer()

    @sub_app.command()
    def child(ctx: typer.Context) -> None:
        root = ctx.find_root()
        assert root.parent is None
        assert root is ctx.parent.parent
        print("ok")

    app.add_typer(sub_app, name="sub")

    result = runner.invoke(app, ["sub", "child"])
    assert result.exit_code == 0
    assert "ok" in result.stdout


def test_context_find_object() -> None:
    class Marker:
        pass

    marker = Marker()
    app = typer.Typer()

    @app.callback()
    def callback(ctx: typer.Context) -> None:
        ctx.obj = marker

    @app.command()
    def child(ctx: typer.Context) -> None:
        assert ctx.find_object(Marker) is marker
        assert ctx.find_object(str) is None
        print("ok")

    result = runner.invoke(app, ["child"])
    assert result.exit_code == 0
    assert "ok" in result.stdout


def test_context_lookup_default_callable() -> None:
    app = typer.Typer()

    @app.command()
    def child(ctx: typer.Context) -> None:
        ctx.default_map = {"planet": lambda: "Earth"}
        assert ctx.lookup_default("planet") == "Earth"
        value = ctx.lookup_default("planet", call=False)
        assert callable(value)
        print("ok")

    result = runner.invoke(app)
    assert result.exit_code == 0
    assert "ok" in result.stdout


def test_context_abort() -> None:
    app = typer.Typer()

    @app.command()
    def cmd(ctx: typer.Context) -> None:
        ctx.abort()

    result = runner.invoke(app, standalone_mode=False)
    assert result.exit_code == 1
    assert isinstance(result.exception, _click.core.Abort)


def test_command_help_disabled() -> None:
    app = typer.Typer()

    @app.command(add_help_option=False)
    def cmd() -> None:
        pass  # pragma: no cover

    result = runner.invoke(app, ["--help"], standalone_mode=False)
    assert result.exit_code == 1
    assert isinstance(result.exception, _click.exceptions.NoSuchOption)
    assert result.exception.option_name == "--help"


def test_command_help_deprecated() -> None:
    app = typer.Typer(rich_markup_mode=None, epilog="Built with love")

    @app.command(short_help="Shorty", help="Regular help text.", deprecated=True)
    def one() -> None:
        pass  # pragma: no cover

    @app.command()
    def two() -> None:
        pass  # pragma: no cover

    result = runner.invoke(app, ["one", "--help"])
    assert result.exit_code == 0
    assert "Regular help text. (DEPRECATED)" in result.output

    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "Built with love" in result.output
    assert "oneShorty(DEPRECATED)" in result.output.replace(" ", "")


@pytest.mark.parametrize(
    ("value", "expected_prefix", "expected_opt"),
    [
        ("--verbose", "--", "verbose"),
        ("//verbose", "//", "verbose"),
        ("-verbose", "-", "verbose"),
        ("verbose", "", "verbose"),
    ],
)
def test_split_opt(value: str, expected_prefix: str, expected_opt: str) -> None:
    prefix, opt = _split_opt(value)
    assert prefix == expected_prefix
    assert opt == expected_opt


def test_nargs_default_map():
    app = typer.Typer()

    @app.command()
    def main(names: list[str] = typer.Option(None)):
        print(names)  # pragma: no cover

    result = runner.invoke(app, [], default_map={"names": "not-a-list"})
    assert result.exit_code == 2
    assert "Invalid value" in result.output


def test_parameter_name_casing():
    app = typer.Typer()

    @app.command()
    def main(
        arg1: int,
        arg2: int = 42,
        arg3: int = typer.Argument(...),
        ARG4: int = typer.Argument(42),
        ARG5: int = typer.Option(...),
        arg6: int = typer.Option(42),
        arg7: int = typer.Argument(42, metavar="meta7"),
        arg8: int = typer.Argument(metavar="ARG8"),
        arg9: int = typer.Option(metavar="ARG9"),
    ):
        print(
            f"arg1={arg1} arg2={arg2} arg3={arg3} ARG4={ARG4} ARG5={ARG5} "
            f"arg6={arg6} arg7={arg7} arg8={arg8} arg9={arg9}"
        )

    result = runner.invoke(
        app,
        [
            "1",
            "3",
            "4",
            "7",
            "8",
            "--arg2",
            "2",
            "--ARG5",
            "5",
            "--arg6",
            "6",
            "--ARG9",
            "9",
        ],
    )
    assert result.exit_code == 0
    assert (
        "arg1=1 arg2=2 arg3=3 ARG4=4 ARG5=5 arg6=6 arg7=7 arg8=8 arg9=9"
        in result.output
    )

    result = runner.invoke(app, ["1", "3", "4", "7", "8", "--ARG5", "5", "--ARG9", "9"])
    assert result.exit_code == 0
    assert (
        "arg1=1 arg2=42 arg3=3 ARG4=4 ARG5=5 arg6=42 arg7=7 arg8=8 arg9=9"
        in result.output
    )

    result = runner.invoke(app, ["1", "3", "4", "7", "8", "--arg5", "5", "--ARG9", "9"])
    assert result.exit_code != 0
    assert "No such option: --arg5" in result.output

    result = runner.invoke(app, ["1", "3", "4", "7", "8", "--ARG5", "5", "--arg9", "9"])
    assert result.exit_code != 0
    assert "No such option: --arg9" in result.output



import time

import typer

app = typer.Typer()


@app.command()
def write_atomic(
    config: typer.FileText = typer.Option(..., mode="w", atomic=True),
    pause: float = typer.Option(0.3),
) -> None:
    config.write("atomic-content-1\n")
    config.flush()
    typer.echo("halfway")
    time.sleep(pause)
    config.write("atomic-content-2\n")
    config.flush()
    typer.echo("written atomically")


@app.command()
def write_atomic_binary(
    config: typer.FileBinaryWrite = typer.Option(..., atomic=True, lazy=False),
) -> None:
    config.write(b"\x00\x01binary-atomic\n")
    typer.echo("written binary atomically")


@app.command()
def api_atomic(
    config: typer.FileText = typer.Option(..., mode="w", atomic=True, lazy=False),
) -> None:
    typer.echo(f"name={config.name}")
    typer.echo(f"repr={repr(config)}")
    with config as entered:
        typer.echo(f"entered={entered is config}")
        entered.write("atomic-api-done\n")


@app.command()
def invalid_atomic_append(
    config: typer.FileText = typer.Option(..., mode="a", atomic=True, lazy=False),
) -> None:
    typer.echo(config.name)  # pragma: no cover


@app.command()
def invalid_atomic_exclusive(
    config: typer.FileText = typer.Option(..., mode="x", atomic=True, lazy=False),
) -> None:
    typer.echo(config.name)  # pragma: no cover


@app.command()
def invalid_atomic_read(
    config: typer.FileText = typer.Option(..., mode="r", atomic=True, lazy=False),
) -> None:
    typer.echo(config.name)  # pragma: no cover


if __name__ == "__main__":
    app()



from enum import Enum

import typer

app = typer.Typer()


class User(str, Enum):
    rick = "rick"
    morty = "morty"


@app.command()
def main(name: User = typer.Option(User.rick, "--name")):
    print(name.value)


if __name__ == "__main__":
    app()



import subprocess
import sys
from pathlib import Path

import pytest

from . import atomic_write_example as mod


def test_atomic_write(tmp_path: Path) -> None:
    original_content = "existing-content\n"
    output_file = tmp_path / "atomic-write-target.txt"
    output_file.write_text(original_content, encoding="utf-8")

    process = subprocess.Popen(
        [
            sys.executable,
            "-m",
            "coverage",
            "run",
            mod.__file__,
            "write-atomic",
            f"--config={output_file}",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    assert process.stdout is not None

    # Halfway of writing the file, check that the original content is still there
    halfway_line = process.stdout.readline().strip()
    assert halfway_line == "halfway"
    assert output_file.read_text(encoding="utf-8") == original_content

    # Only at the end, the full new content is visible
    stdout, stderr = process.communicate(timeout=5)
    assert process.returncode == 0, stderr
    assert "written atomically" in stdout
    assert (
        output_file.read_text(encoding="utf-8")
        == "atomic-content-1\natomic-content-2\n"
    )


def test_atomic_binary_write(tmp_path: Path) -> None:
    output_file = tmp_path / "atomic-binary.bin"

    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "coverage",
            "run",
            mod.__file__,
            "write-atomic-binary",
            f"--config={output_file}",
        ],
        capture_output=True,
        encoding="utf-8",
    )

    assert result.returncode == 0, result.stderr
    assert "written binary atomically" in result.stdout
    assert output_file.read_bytes() == b"\x00\x01binary-atomic\n"


def test_atomic_api(tmp_path: Path) -> None:
    output_file = tmp_path / "atomic-api.txt"

    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "coverage",
            "run",
            mod.__file__,
            "api-atomic",
            f"--config={output_file}",
        ],
        capture_output=True,
        encoding="utf-8",
    )

    assert result.returncode == 0, result.stderr
    assert f"name={output_file}" in result.stdout
    assert "repr=<_io.TextIOWrapper" in result.stdout
    assert "entered=True" in result.stdout
    assert output_file.read_text(encoding="utf-8") == "atomic-api-done\n"


@pytest.mark.parametrize(
    ("command_name", "expected_message"),
    [
        ("invalid-atomic-append", "Appending to an existing file is not supported"),
        ("invalid-atomic-exclusive", "Use the `overwrite`-parameter instead."),
        ("invalid-atomic-read", "Atomic writes only make sense with `w`-mode."),
    ],
)
def test_atomic_mode_invalid_options(
    tmp_path: Path, command_name: str, expected_message: str
) -> None:
    output_file = tmp_path / "atomic-invalid-mode.txt"
    output_file.write_text("existing-content\n", encoding="utf-8")

    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "coverage",
            "run",
            mod.__file__,
            command_name,
            f"--config={output_file}",
        ],
        capture_output=True,
        encoding="utf-8",
    )

    assert result.returncode != 0
    combined_output = f"{result.stdout}\n{result.stderr}"
    assert expected_message in combined_output



import io
import subprocess
from unittest.mock import Mock, patch

import pytest
import typer

from tests.utils import needs_linux, needs_macos, needs_windows

url = "http://example.com"


@pytest.mark.parametrize(
    "system, command",
    [
        ("Darwin", "open"),
        ("Linux", "xdg-open"),
        ("FreeBSD", "xdg-open"),
    ],
)
def test_launch_url_unix(system: str, command: str):
    with (
        patch("platform.system", return_value=system),
        patch("shutil.which", return_value=True),
        patch("subprocess.Popen") as mock_popen,
    ):
        typer.launch(url)

    mock_popen.assert_called_once_with(
        [command, url], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
    )


def test_launch_url_windows():
    with (
        patch("platform.system", return_value="Windows"),
        patch("webbrowser.open") as mock_webbrowser_open,
    ):
        typer.launch(url)

    mock_webbrowser_open.assert_called_once_with(url)


def test_launch_url_no_xdg_open():
    with (
        patch("platform.system", return_value="Linux"),
        patch("shutil.which", return_value=None),
        patch("webbrowser.open") as mock_webbrowser_open,
    ):
        typer.launch(url)

    mock_webbrowser_open.assert_called_once_with(url)


def test_launch_url_linux_xdg_open_does_not_wait_by_default():
    proc = Mock()

    with (
        patch("platform.system", return_value="Linux"),
        patch("shutil.which", return_value=True),
        patch("subprocess.Popen", return_value=proc) as mock_popen,
    ):
        result = typer.launch(url)

    assert result == 0
    mock_popen.assert_called_once_with(
        ["xdg-open", url], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
    )
    proc.wait.assert_not_called()


def test_launch_url_linux_xdg_open_waits_when_requested():
    proc = Mock()
    proc.wait.return_value = 42

    with (
        patch("platform.system", return_value="Linux"),
        patch("shutil.which", return_value=True),
        patch("subprocess.Popen", return_value=proc) as mock_popen,
    ):
        result = typer.launch(url, wait=True)

    assert result == 42
    mock_popen.assert_called_once_with(
        ["xdg-open", url], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
    )
    proc.wait.assert_called_once_with()


@pytest.fixture
def allow_dev_null(monkeypatch):
    real_open = open

    def fake_open(path, *args, **kwargs):
        if path == "/dev/null":
            return io.StringIO()
        return real_open(path, *args, **kwargs)  # pragma: no cover

    monkeypatch.setattr("builtins.open", fake_open)


@needs_macos
def test_open_url_macos(monkeypatch, allow_dev_null):
    recorded: list[list[str]] = []

    class Proc:
        def wait(self) -> int:
            return 42

    def fake_popen(args, **kwargs):
        recorded.append(list(args))
        return Proc()

    monkeypatch.setattr(subprocess, "Popen", fake_popen)

    assert typer.launch("/path/to/file", wait=True, locate=True) == 42
    assert recorded[0][:3] == ["open", "-W", "-R"]
    assert recorded[0][-1] == "/path/to/file"


@needs_windows
def test_launch_files_windows(monkeypatch):
    calls: list[list[str]] = []

    def fake_call(args):
        calls.append(list(args))
        return 0

    monkeypatch.setattr(subprocess, "call", fake_call)

    assert typer.launch("C:/Tools/readme.txt", wait=True, locate=False) == 0
    assert typer.launch("file:///C:/tmp/a.txt", wait=False, locate=True) == 0
    assert calls.pop(0) == ["start", "/WAIT", "", "C:/Tools/readme.txt"]
    assert calls.pop(0) == ["explorer", "/select,/C:/tmp/a.txt"]

    monkeypatch.setattr(subprocess, "call", lambda a: (_ for _ in ()).throw(OSError()))
    assert typer.launch("D:/no/such/file.txt", wait=False, locate=False) == 127


@needs_linux
def test_open_url_linux_wait(monkeypatch):
    class Proc:
        def __init__(self, code: int = 0) -> None:
            self._code = code

        def wait(self) -> int:
            return self._code

    monkeypatch.setattr(subprocess, "Popen", lambda *a, **k: Proc(7))

    assert typer.launch("/file", wait=True, locate=False) == 7


@needs_linux
def test_open_url_linux_locate(monkeypatch):
    recorded: list[list[str]] = []

    class Proc:
        def wait(self) -> int:
            return 0  # pragma: no cover

    def fake_popen(args, **kwargs):
        recorded.append(list(args))
        return Proc()

    monkeypatch.setattr(subprocess, "Popen", fake_popen)

    assert typer.launch("/tmp/sub/file.txt", wait=False, locate=True) == 0
    assert recorded[-1] == ["xdg-open", "/tmp/sub"]


@needs_windows
def test_launch_file():
    with (
        patch("typer._click._termui_impl.sys.platform", "win32"),
        patch("typer._click._termui_impl.WIN", True),
        patch("typer._click._termui_impl.CYGWIN", False),
        patch("subprocess.call", return_value=0) as call_mock,
    ):
        result = typer.launch("C:/tmp/file.txt", locate=True)

    assert result == 0
    call_mock.assert_called_once_with(["explorer", "/select,C:/tmp/file.txt"])



"""
Tests for the termui, echo, and CliRunner isolation functionality.
Created after vendoring Click to ensure test coverage is back up to 100%.
"""

import io
import os
from contextlib import contextmanager
from typing import Literal

import pytest
import typer
from typer._click import _termui_impl, termui
from typer.testing import CliRunner

from tests.utils import needs_windows, skip_if_windows


def test_raw_terminal(monkeypatch):
    runner = CliRunner()
    app = typer.Typer()
    state = {"entered": 0, "exited": 0}

    @contextmanager
    def fake_raw_terminal():
        state["entered"] += 1
        try:
            yield 42
        finally:
            state["exited"] += 1

    monkeypatch.setattr(_termui_impl, "raw_terminal", fake_raw_terminal)

    @app.command()
    def main():
        with termui.raw_terminal() as fd:
            typer.echo(f"fd={fd}")

    result = runner.invoke(app, [])
    assert result.exit_code == 0, result.output
    assert "fd=42" in result.stdout
    assert state["entered"] == 1
    assert state["exited"] == 1


def test_getchar(monkeypatch):
    # Cached path: call the existing _getchar directly.
    cached_state = {"echo": None}

    def cached_getchar(echo: bool) -> str:
        cached_state["echo"] = echo
        return "x"

    monkeypatch.setattr(termui, "_getchar", cached_getchar)
    assert termui.getchar(echo=True) == "x"
    assert cached_state["echo"] is True

    # Lazy-load path: _getchar is None, so import/cache _termui_impl.getchar.
    lazy_state = {"calls": 0}

    def lazy_getchar(echo: bool) -> str:
        lazy_state["calls"] += 1
        return "y" if not echo else "z"

    monkeypatch.setattr(termui, "_getchar", None)
    monkeypatch.setattr(_termui_impl, "getchar", lazy_getchar)

    assert termui.getchar(echo=False) == "y"
    assert termui._getchar is lazy_getchar
    assert termui.getchar(echo=True) == "z"
    assert lazy_state["calls"] == 2


def test_clirunner_getchar(monkeypatch) -> None:
    runner = CliRunner()
    app = typer.Typer()

    @app.command()
    def main() -> None:
        first = termui.getchar(echo=False)
        second = termui.getchar(echo=True)
        typer.echo(f"\nfirst={first};second={second}")

    monkeypatch.setattr(termui, "_getchar", None)
    result = runner.invoke(app, [], input="ab")
    assert result.exit_code == 0, result.output
    assert result.stdout.splitlines() == ["b", "first=a;second=b"]


def test_clirunner_env_none(monkeypatch) -> None:
    runner = CliRunner()
    app = typer.Typer()
    env_key = "TYPER_TEST_ENV_REMOVE"
    monkeypatch.setenv(env_key, "present")

    @app.command()
    def main() -> None:
        typer.echo(f"inside={os.environ.get(env_key)}")

    result = runner.invoke(app, [], env={env_key: None})
    assert result.exit_code == 0, result.output
    assert "inside=None" in result.stdout
    assert os.environ.get(env_key) == "present"


@pytest.mark.parametrize(
    ("exit_value", "expected_exit_code", "expected_stdout"),
    [
        (None, 0, ""),
        ("bad-exit", 1, "bad-exit\n"),
    ],
)
def test_clirunner_invoke_system_exit_branches(
    exit_value: object,
    expected_exit_code: int,
    expected_stdout: str,
) -> None:
    runner = CliRunner()
    app = typer.Typer()

    @app.command()
    def main() -> None:
        raise SystemExit(exit_value)

    result = runner.invoke(app, [])
    assert result.exit_code == expected_exit_code
    assert result.stdout == expected_stdout
    if expected_exit_code:
        assert isinstance(result.exception, SystemExit)
    else:
        assert result.exception is None


@needs_windows
def test_termui_impl_windows_raw_terminal():
    with _termui_impl.raw_terminal() as fd:
        assert fd == -1
    with termui.raw_terminal() as fd:
        assert fd == -1


@needs_windows
def test_termui_impl_windows_getchar(monkeypatch):
    monkeypatch.setattr(_termui_impl.msvcrt, "getwch", lambda: "a")
    monkeypatch.setattr(_termui_impl.msvcrt, "getwche", lambda: "b")
    assert _termui_impl.getchar(echo=False) == "a"
    assert _termui_impl.getchar(echo=True) == "b"

    seq_null = iter(["\x00", "K"])
    monkeypatch.setattr(_termui_impl.msvcrt, "getwch", lambda: next(seq_null))
    assert _termui_impl.getchar(echo=False) == "\x00K"

    seq_e0 = iter(["\xe0", "H"])
    monkeypatch.setattr(_termui_impl.msvcrt, "getwch", lambda: next(seq_e0))
    assert _termui_impl.getchar(echo=False) == "\xe0H"

    seq_echo = iter(["\x00", "M"])
    monkeypatch.setattr(_termui_impl.msvcrt, "getwche", lambda: next(seq_echo))
    assert _termui_impl.getchar(echo=True) == "\x00M"

    seq_e0_echo = iter(["\xe0", "Z"])
    monkeypatch.setattr(_termui_impl.msvcrt, "getwche", lambda: next(seq_e0_echo))
    assert _termui_impl.getchar(echo=True) == "\xe0Z"

    monkeypatch.setattr(_termui_impl.msvcrt, "getwch", lambda: "\x03")
    with pytest.raises(KeyboardInterrupt):
        _termui_impl.getchar(echo=False)

    monkeypatch.setattr(_termui_impl.msvcrt, "getwch", lambda: "\x1a")
    with pytest.raises(EOFError):
        _termui_impl.getchar(echo=False)


@skip_if_windows
@pytest.mark.parametrize("use_stdin_tty", [True, False])
def test_termui_impl_posix_raw_terminal(monkeypatch, use_stdin_tty: bool):
    state: dict[str, object] = {}
    flushed: list[None] = []
    fake_tty = None

    if use_stdin_tty:
        expected_fd = 14
        old_termios = "old_settings"
        monkeypatch.setattr(
            _termui_impl, "isatty", lambda s: s is _termui_impl.sys.stdin
        )
        monkeypatch.setattr(_termui_impl.sys.stdin, "fileno", lambda: expected_fd)
    else:
        expected_fd = 27
        old_termios = "old"
        monkeypatch.setattr(
            _termui_impl,
            "isatty",
            lambda s: s is not _termui_impl.sys.stdin,
        )

        class FakeTTY:
            def __init__(self) -> None:
                self.closed = False

            def fileno(self) -> int:
                return expected_fd

            def close(self) -> None:
                self.closed = True

        fake_tty = FakeTTY()
        real_open = open

        def fake_open(path, *args, **kwargs):
            if path == "/dev/tty":
                return fake_tty
            return real_open(path, *args, **kwargs)  # pragma: no cover

        monkeypatch.setattr("builtins.open", fake_open)

    def tcgetattr(fd: int) -> str:
        state["tcgetattr_fd"] = fd
        return old_termios

    def setraw(fd: int) -> None:
        state["setraw_fd"] = fd

    def tcsetattr(fd: int, when: int, old: str) -> None:
        state["tcsetattr"] = (fd, when, old)

    monkeypatch.setattr(_termui_impl.termios, "tcgetattr", tcgetattr)
    monkeypatch.setattr(_termui_impl.tty, "setraw", setraw)
    monkeypatch.setattr(_termui_impl.termios, "tcsetattr", tcsetattr)
    monkeypatch.setattr(
        _termui_impl.sys.stdout, "flush", lambda *a, **k: flushed.append(None)
    )

    with _termui_impl.raw_terminal() as fd:
        assert fd == expected_fd

    assert state["tcgetattr_fd"] == expected_fd
    assert state["setraw_fd"] == expected_fd
    assert state["tcsetattr"] == (
        expected_fd,
        _termui_impl.termios.TCSADRAIN,
        old_termios,
    )
    assert flushed == [None]
    if fake_tty is not None:
        assert fake_tty.closed is True


@skip_if_windows
def test_termui_impl_posix_getchar(monkeypatch):
    @contextmanager
    def fake_raw():
        yield 7

    monkeypatch.setattr(_termui_impl, "raw_terminal", fake_raw)
    monkeypatch.setattr(_termui_impl.os, "read", lambda fd, n: b"q")
    monkeypatch.setattr(_termui_impl, "get_best_encoding", lambda stdin: "utf-8")
    monkeypatch.setattr(_termui_impl, "isatty", lambda f: f is _termui_impl.sys.stdout)
    written: list[str] = []
    monkeypatch.setattr(_termui_impl.sys.stdout, "write", lambda s: written.append(s))

    assert _termui_impl.getchar(echo=True) == "q"
    assert written == ["q"]


@skip_if_windows
def test_termui_impl_posix_getchar_eof(monkeypatch):
    @contextmanager
    def fake_raw():
        yield 5

    monkeypatch.setattr(_termui_impl, "raw_terminal", fake_raw)
    monkeypatch.setattr(_termui_impl.os, "read", lambda fd, n: b"\x04")
    monkeypatch.setattr(_termui_impl, "get_best_encoding", lambda stdin: "utf-8")
    monkeypatch.setattr(_termui_impl, "isatty", lambda f: False)

    with pytest.raises(EOFError):
        _termui_impl.getchar(echo=False)


def test_prompt():
    runner = CliRunner()
    app = typer.Typer()
    fake_file = io.StringIO("data")
    fake_file.name = "demo.txt"

    @app.command()
    def main(
        accept: bool = typer.Option(True, prompt=True),
        name: str = typer.Option(..., prompt=True),
        flavor: Literal["a", "b"] = typer.Option(..., prompt=True),
        city: str = typer.Option("London", prompt=True),
        config: str = typer.Option(fake_file, prompt=True),
        password: str = typer.Option(
            ...,
            prompt=True,
            hide_input=True,
            confirmation_prompt=True,
        ),
    ):
        typer.echo(
            f"accept={accept};name={name};flavor={flavor};city={city};config={config};pass_len={len(password)}"
        )

    result = runner.invoke(app, [], input="\nAda\na\n\ncustom.ini\nsecret\nsecret\n")
    assert result.exit_code == 0, result.output
    assert (
        "accept=True;name=Ada;flavor=a;city=London;config=custom.ini;pass_len=6"
        in result.stdout
    )
    assert "(a, b): " in result.stdout
    assert "[demo.txt]: " in result.stdout


def test_hidden_prompt_func(monkeypatch):
    monkeypatch.setattr("getpass.getpass", lambda prompt: "secret")
    assert termui.hidden_prompt_func("Password: ") == "secret"


def test_echo_stdout_missing(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("sys.stdout", None)
    typer.echo("ignored")


def test_echo_stringifies() -> None:
    stream = io.StringIO()
    typer.echo(123, file=stream, nl=False)
    assert stream.getvalue() == "123"


def test_echo_bytes() -> None:
    buffer = io.BytesIO()
    stream = io.TextIOWrapper(buffer, encoding="utf-8")
    typer.echo(b"abc", file=stream, nl=True)
    assert buffer.getvalue() == b"abc\n"


def test_echo_empty_output() -> None:
    class FlushTrackingTextStream(io.StringIO):
        def __init__(self) -> None:
            super().__init__()
            self.flush_count = 0

        def flush(self) -> None:
            self.flush_count += 1
            super().flush()

        def write(self, s: str) -> int:
            raise AssertionError("Empty output")  # pragma: no cover

    stream = FlushTrackingTextStream()
    typer.echo("", file=stream, nl=False)
    assert stream.flush_count == 1


@needs_windows
def test_echo_windows_color_none(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    class TtyStream(io.StringIO):
        def isatty(self) -> bool:
            return True

    stream = TtyStream()
    monkeypatch.setattr("typer._click.utils.auto_wrap_for_ansi", None)
    typer.echo("\x1b[31mred\x1b[0m", file=stream, nl=False, color=None)
    assert stream.getvalue() == "red"


@pytest.mark.parametrize(
    ("flag", "true_code", "false_code"),
    [
        ("bold", "\x1b[1m", "\x1b[22m"),
        ("dim", "\x1b[2m", "\x1b[22m"),
        ("underline", "\x1b[4m", "\x1b[24m"),
        ("overline", "\x1b[53m", "\x1b[55m"),
        ("italic", "\x1b[3m", "\x1b[23m"),
        ("blink", "\x1b[5m", "\x1b[25m"),
        ("reverse", "\x1b[7m", "\x1b[27m"),
        ("strikethrough", "\x1b[9m", "\x1b[29m"),
    ],
)
def test_style(flag, true_code, false_code):
    runner = CliRunner()
    app = typer.Typer()

    @app.command()
    def main():
        # testing an int and a str on purpose
        typer.echo("TRUE=" + typer.style("42", **{flag: True}), color=True)
        typer.echo("FALSE=" + typer.style(666, **{flag: False}), color=True)

    result = runner.invoke(app, [])
    assert result.exit_code == 0, result.output
    lines = [line for line in result.stdout.splitlines() if line]
    true_line = next(line for line in lines if line.startswith("TRUE="))
    false_line = next(line for line in lines if line.startswith("FALSE="))
    assert "42" in true_line
    assert "666" in false_line

    assert true_code in true_line
    assert true_code not in false_line
    assert false_code in false_line
    assert false_code not in true_line


def test_style_color():
    fg_int = typer.style("x", fg=123)
    assert "\x1b[38;5;123m" in fg_int

    bg_list = typer.style("x", bg=[1, 2, 3])
    assert "\x1b[48;2;1;2;3m" in bg_list

    with pytest.raises(TypeError, match="Unknown color"):
        typer.style("x", fg="not-a-color")

    with pytest.raises(TypeError, match="Unknown color"):
        typer.style("x", bg="not-a-color")


def test_termui_launch(monkeypatch):
    captured = {}

    def fake_open_url(url, wait=False, locate=False):
        captured["url"] = url
        captured["wait"] = wait
        captured["locate"] = locate
        return 7

    monkeypatch.setattr(_termui_impl, "open_url", fake_open_url)
    rv = termui.launch("https://example.com", wait=True, locate=True)
    assert rv == 7
    assert captured == {
        "url": "https://example.com",
        "wait": True,
        "locate": True,
    }



import pytest
import typer.core
from typer.testing import CliRunner

from tests.assets import hidden_commands as mod_command
from tests.assets import hidden_option as mod_option
from tests.utils import needs_rich

runner = CliRunner()


@pytest.mark.parametrize(
    "use_rich",
    [
        pytest.param(False),
        pytest.param(True, marks=needs_rich),
    ],
)
def test_hidden_option(monkeypatch: pytest.MonkeyPatch, use_rich: bool) -> None:
    monkeypatch.setattr(typer.core, "HAS_RICH", use_rich)

    result = runner.invoke(mod_option.app, ["--help"])
    assert result.exit_code == 0
    assert "Say hello" in result.output
    assert "--name" not in result.output
    assert "/lastname" in result.output
    assert "TEST_LASTNAME" in result.output
    assert "(dynamic)" in result.output


def test_coverage_call() -> None:
    result = runner.invoke(mod_option.app)
    assert result.exit_code == 0
    assert "Hello John Doe, it seems you have 42" in result.output


@pytest.mark.parametrize(
    "use_rich",
    [
        pytest.param(False),
        pytest.param(True, marks=needs_rich),
    ],
)
def test_hidden_commands(monkeypatch: pytest.MonkeyPatch, use_rich: bool) -> None:
    monkeypatch.setattr(typer.core, "HAS_RICH", use_rich)

    result = runner.invoke(mod_command.app, ["--help"])
    assert result.exit_code == 0
    assert "visible" in result.output
    assert "hidden-decorated" not in result.output
    assert "hidden-var" not in result.output
    assert "hidden-subgroup" not in result.output



import subprocess
import sys

import pytest
import typer
import typer.core
from typer.testing import CliRunner

from docs_src.arguments.optional import tutorial003_py310 as mod

runner = CliRunner()
app = mod.app


def test_call_no_arg():
    result = runner.invoke(app)
    assert result.exit_code != 0
    assert "Missing argument 'name'." in result.output


def test_call_no_arg_standalone():
    # Mainly for coverage
    result = runner.invoke(app, standalone_mode=False)
    assert result.exit_code != 0


def test_call_no_arg_no_rich(monkeypatch: pytest.MonkeyPatch):
    # Mainly for coverage
    monkeypatch.setattr(typer.core, "HAS_RICH", False)
    result = runner.invoke(app)
    assert result.exit_code != 0
    assert "Error: Missing argument 'name'" in result.output


def test_call_arg():
    result = runner.invoke(app, ["Camila"])
    assert result.exit_code == 0
    assert "Hello Camila" in result.output


def test_script():
    result = subprocess.run(
        [sys.executable, "-m", "coverage", "run", mod.__file__, "--help"],
        capture_output=True,
        encoding="utf-8",
    )
    assert "Usage" in result.stdout



import typer

app = typer.Typer()


def complete_name(ctx, args, incomplete):
    return ["opt-choice"]  # pragma: no cover


def complete_target(ctx, args, incomplete):
    return ["arg-choice"]


@app.command()
def main(
    name: str = typer.Option(..., "--name", autocompletion=complete_name),
    target: str = typer.Argument(..., autocompletion=complete_target),
):
    print(name, target)  # pragma: no cover


if __name__ == "__main__":
    app()



import typer
import typer.main

app = typer.Typer()


@app.command()
def main(name: str = "morty"):
    # The line below will cause TypeError because you cannot concatenate string and integer.
    print(name + 3)


if __name__ == "__main__":
    app()



"""Prepare a release by updating the package version and release notes."""

import re
from datetime import date
from pathlib import Path
from typing import Annotated, Literal

import typer

VERSION_PATTERN = re.compile(r'(?m)^__version__ = "(\d+\.\d+\.\d+)"$')
VERSION_HEADING_PATTERN = re.compile(r"(?m)^## (\d+\.\d+\.\d+)(?: \([^)]+\))?$")
RELEASE_NOTES_HEADER = "# Release Notes\n\n"
LATEST_CHANGES_HEADER = "## Latest Changes"
BumpType = Literal["major", "minor", "patch"]

app = typer.Typer()


def parse_version(version: str) -> tuple[int, int, int]:
    match = re.fullmatch(r"\d+\.\d+\.\d+", version)
    if not match:
        raise ValueError(f"Invalid version: {version!r}. Expected format: X.Y.Z")
    major, minor, patch = version.split(".")
    return int(major), int(minor), int(patch)


def get_current_version(content: str, version_file: Path) -> str:
    matches = list(VERSION_PATTERN.finditer(content))
    if len(matches) != 1:
        raise RuntimeError(
            f"Expected exactly one __version__ assignment in {version_file}, "
            f"found {len(matches)}"
        )
    return matches[0].group(1)


def bump_version(version: str, bump: BumpType) -> str:
    major, minor, patch = parse_version(version)
    if bump == "major":
        return f"{major + 1}.0.0"
    if bump == "minor":
        return f"{major}.{minor + 1}.0"
    return f"{major}.{minor}.{patch + 1}"


def update_version_file(content: str, version: str, version_file: Path) -> str:
    current_version = get_current_version(content, version_file)
    if parse_version(version) <= parse_version(current_version):
        raise RuntimeError(
            f"New version {version} must be greater than current version {current_version}"
        )
    return VERSION_PATTERN.sub(f'__version__ = "{version}"', content, count=1)


def update_release_notes(
    content: str, version: str, release_date: date, release_notes_file: Path
) -> str:
    if not content.startswith(RELEASE_NOTES_HEADER):
        raise RuntimeError(
            f"{release_notes_file} must start with {RELEASE_NOTES_HEADER!r}"
        )
    if re.search(rf"^## {re.escape(version)}(?: \([^)]+\))?$", content, re.M):
        raise RuntimeError(f"Release notes already contain a section for {version}")

    latest_header = f"{RELEASE_NOTES_HEADER}{LATEST_CHANGES_HEADER}\n"
    if not content.startswith(latest_header):
        raise RuntimeError(f"{release_notes_file} must start with {latest_header!r}")

    release_header = f"## {version} ({release_date.isoformat()})"
    return content.replace(
        latest_header,
        f"{RELEASE_NOTES_HEADER}{LATEST_CHANGES_HEADER}\n\n{release_header}\n",
        1,
    )


def get_release_notes_body(content: str, version: str, release_notes_file: Path) -> str:
    version_heading = re.compile(rf"(?m)^## {re.escape(version)}(?: \([^)]+\))?$")
    match = version_heading.search(content)
    if not match:
        raise RuntimeError(
            f"Could not find release notes section for {version} in {release_notes_file}"
        )

    next_match = VERSION_HEADING_PATTERN.search(content, match.end())
    end = next_match.start() if next_match else len(content)
    body = content[match.end() : end].strip()
    if not body:
        raise RuntimeError(
            f"Release notes section for {version} in {release_notes_file} is empty"
        )
    return f"{body}\n"


@app.command()
def prepare(
    bump: Annotated[
        BumpType,
        typer.Argument(
            envvar="PREPARE_RELEASE_BUMP",
            help="The release bump to make: major, minor, or patch.",
        ),
    ],
    version_file: Annotated[
        Path,
        typer.Option(
            envvar="PREPARE_RELEASE_VERSION_FILE",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            writable=True,
            help="Path to the Python file containing the __version__ assignment.",
        ),
    ],
    release_notes_file: Annotated[
        Path,
        typer.Option(
            envvar="PREPARE_RELEASE_RELEASE_NOTES_FILE",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            writable=True,
            help="Path to the release notes Markdown file.",
        ),
    ],
    release_date: Annotated[
        str,
        typer.Option(
            "--date",
            envvar="PREPARE_RELEASE_DATE",
            help="Release date in YYYY-MM-DD format. Defaults to today.",
        ),
    ] = date.today().isoformat(),
) -> None:
    parsed_release_date = date.fromisoformat(release_date or date.today().isoformat())

    version_file_content = version_file.read_text()
    release_notes_content = release_notes_file.read_text()
    version = bump_version(
        get_current_version(version_file_content, version_file), bump
    )

    version_file.write_text(
        update_version_file(version_file_content, version, version_file)
    )
    release_notes_file.write_text(
        update_release_notes(
            release_notes_content, version, parsed_release_date, release_notes_file
        )
    )

    typer.echo(f"Prepared release {version} ({parsed_release_date.isoformat()})")


@app.command()
def current_version(
    version_file: Annotated[
        Path,
        typer.Option(
            envvar="PREPARE_RELEASE_VERSION_FILE",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="Path to the Python file containing the __version__ assignment.",
        ),
    ],
) -> None:
    typer.echo(get_current_version(version_file.read_text(), version_file))


@app.command()
def release_notes(
    version_file: Annotated[
        Path,
        typer.Option(
            envvar="PREPARE_RELEASE_VERSION_FILE",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="Path to the Python file containing the __version__ assignment.",
        ),
    ],
    release_notes_file: Annotated[
        Path,
        typer.Option(
            envvar="PREPARE_RELEASE_RELEASE_NOTES_FILE",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="Path to the release notes Markdown file.",
        ),
    ],
) -> None:
    version = get_current_version(version_file.read_text(), version_file)
    typer.echo(
        get_release_notes_body(
            release_notes_file.read_text(), version, release_notes_file
        ),
        nl=False,
    )


if __name__ == "__main__":
    app()
